from typing import Optional

import numpy as np

from dtaianomaly import utils
from dtaianomaly.anomaly_detection.BaseDetector import BaseDetector, Supervision


class MedianMethod(BaseDetector):
    """
    Anomaly detection based on the Two-sided Median Method.

    The Median Method [basu2007automatic]_ computes the deviation of a sample
    compared to its neighborhood. This neighborhood is computed as a window
    around the sample. The deviation is consequently measured as the number
    of standard deviations     the observations deviates from the mean of its
    neighborhood.

    In contrast to the original paper, this implementation allows to define a
    different neighborhood size before and after the sample, to fine tune how
    much lookahead is allowed. In the ultimate case, if ``neighborhood_size_after = 0``,
    then the Median Method is a purely online anomaly detector. Note, however,
    that this case differs from the One-Sided Median Method discussed in the
    original paper, which also uses the first order difference to detect anomalies.

    Parameters
    ----------
    neighborhood_size_before: int
        The number of observations before the sample to include in the neighborhood.
    neighborhood_size_after: int, default=None
        The number of observations after the sample to include in the neighborhood.
        If None, the same value as ``window_size_before`` will be used.

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import MedianMethod
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> median_method = MedianMethod(10).fit(x)
    >>> median_method.decision_function(x)
    array([1.1851476 , 0.68191703, 1.05125284, ..., 0.81373386, 1.86097851,
           0.05852008])

    References
    ----------
    .. [basu2007automatic] Basu, Sabyasachi, and Martin Meckesheimer. "Automatic outlier
       detection for time series: an application to sensor data." Knowledge and Information
       Systems 11 (2007), 137-154, doi: `10.1007/s10115-006-0026-6 <https://doi.org/10.1007/s10115-006-0026-6>`_.
    """

    neighborhood_size_before: int
    neighborhood_size_after: Optional[int]

    def __init__(
        self,
        neighborhood_size_before: int,
        neighborhood_size_after: Optional[int] = None,
    ):
        super().__init__(Supervision.UNSUPERVISED)

        if not isinstance(neighborhood_size_before, int) or isinstance(
            neighborhood_size_before, bool
        ):
            raise TypeError("`neighborhood_size_before` should be an integer")
        if neighborhood_size_before < 1:
            raise ValueError("`neighborhood_size_before` should be strictly positive")

        if neighborhood_size_after is not None:
            if not isinstance(neighborhood_size_after, int) or isinstance(
                neighborhood_size_after, bool
            ):
                raise TypeError("`neighborhood_size_after` should be an integer")
            if neighborhood_size_after < 0:
                raise ValueError("`neighborhood_size_after` can not be negative!")

        self.neighborhood_size_before = neighborhood_size_before
        self.neighborhood_size_after = neighborhood_size_after

    def fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None, **kwargs
    ) -> "BaseDetector":
        return self  # No fitting is required.

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")
        if not utils.is_univariate(X):
            raise ValueError("Input must be univariate!")

        # Convert to a numpy array
        X = np.asarray(X, dtype=float).squeeze()

        # Set the neighborhood size after the observation
        if self.neighborhood_size_after is None:
            neighborhood_size_after = self.neighborhood_size_before
        else:
            neighborhood_size_after = self.neighborhood_size_after

        X_padded = np.pad(
            X,
            (self.neighborhood_size_before, neighborhood_size_after),
            constant_values=(np.nan,),
        )
        neighborhoods = np.lib.stride_tricks.sliding_window_view(
            X_padded,
            window_shape=(self.neighborhood_size_before + neighborhood_size_after + 1),
        )
        return np.nan_to_num(
            np.abs(X - np.nanmean(neighborhoods, axis=1))
            / np.nanstd(neighborhoods, axis=1),
            nan=0.0,
        )



from typing import List, Optional, Tuple, Union

import numpy as np

from dtaianomaly import utils
from dtaianomaly.preprocessing.Preprocessor import (
    Preprocessor,
    check_preprocessing_inputs,
)


class ChainedPreprocessor(Preprocessor):
    """
    Wrapper chaining multiple `Preprocessor` objects.

    Parameters
    ----------
    base_preprocessors: list of `Preprocessor` objects
        The preprocessors to chain. These preprocessors can be passed as a single
        list argument or as multiple independent arguments to the constructor.
    """

    base_preprocessors: List[Preprocessor]

    def __init__(self, *base_preprocessors: Union[Preprocessor, List[Preprocessor]]):
        # Format the base processors
        if len(base_preprocessors) == 1 and isinstance(base_preprocessors[0], list):
            base_preprocessors = base_preprocessors[0]
        else:
            base_preprocessors = list(base_preprocessors)
        # Check the preprocessors
        if (
            not utils.is_valid_list(base_preprocessors, Preprocessor)
            or len(base_preprocessors) == 0
        ):
            raise ValueError("Expected a list of Preprocessors")
        self.base_preprocessors = base_preprocessors

    def _fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "Preprocessor":
        for preprocessor in self.base_preprocessors:
            preprocessor._fit(X, y)
            X, y = preprocessor._transform(X, y)
        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        for preprocessor in self.base_preprocessors:
            X, y = preprocessor._transform(X, y)
        return X, y

    def fit_transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        check_preprocessing_inputs(X, y)
        for preprocessor in self.base_preprocessors:
            preprocessor._fit(X, y)
            X, y = preprocessor._transform(X, y)
        return X, y

    def __str__(self):
        return "->".join(map(str, self.base_preprocessors))



from pyod.models.pca import PCA

from dtaianomaly.anomaly_detection.BaseDetector import Supervision
from dtaianomaly.anomaly_detection.PyODAnomalyDetector import PyODAnomalyDetector


class PrincipalComponentAnalysis(PyODAnomalyDetector):
    """
    Anomaly detector based on the Principal Component Analysis (PCA).

    PCA [Charu2015outlier]_ maps the data to a lower dimensional space
    through linear projections. The goal of these projections is to
    capture the most important information of the samples. This important
    information is related to the type of behaviors that occur frequently
    in the data. Thus, anomalies are detected by measuring the deviation
    of the samples in the lower dimensional space.

    Notes
    -----
    PCA inherets from :py:class:`~dtaianomaly.anomaly_detection.PyodAnomalyDetector`.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs:
        Arguments to be passed to the PyOD PCA.

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : PCA
        A PCA-detector of PyOD

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import PrincipalComponentAnalysis
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> pca = PrincipalComponentAnalysis(10).fit(x)
    >>> pca.decision_function(x)
    array([16286.63724327, 15951.05917741, 15613.5739773 , ...,
           18596.5273311 , 18496.96613747, 18483.47985554])

    References
    ----------
    .. [Charu2015outlier] Charu C Aggarwal. Outlier analysis. In Data mining, 75–79. Springer, 2015.
    """

    def _initialize_detector(self, **kwargs) -> PCA:
        return PCA(**kwargs)

    def _supervision(self):
        return Supervision.SEMI_SUPERVISED



""" This function is adapted from TSB-AD """

from typing import Optional, Union

import numpy as np
from sklearn.decomposition import PCA
from sklearn.exceptions import NotFittedError

from dtaianomaly import utils
from dtaianomaly.anomaly_detection.BaseDetector import BaseDetector, Supervision
from dtaianomaly.anomaly_detection.windowing_utils import (
    check_is_valid_window_size,
    compute_window_size,
    reverse_sliding_window,
    sliding_window,
)


class RobustPrincipalComponentAnalysis(BaseDetector):
    """
    Anomaly detection based on Robust Principal Component Analysis (Robust PCA).

    Assume that the data matrix is a superposition of a low-rank component and a s
    parse component. Robust PCA [Candes2011robust]_ will solve this decomposition
    as a convex optimization problem. The superposition offers a principeled manner
    to robust PCA, since the methodology can recover the principal components (first
    component) of a data matrix even though a positive fraction of the entries are
    arbitrarly corrupted or anomalous (second component).

    Warnings
    --------
    During testing, we found that there are some deviations in the predicted decision
    scores, depending on if the method was run on windows or linux. The difference in
    the absolute value is of around the order of 2%, but the general trend of the
    anomaly scores remains consistent. The only randomness in this implementation of
    Robust PCA is the PCA solver of scikit-learn, but even setting a random state
    did not resolve the issue.

    Notes
    -----
    In most existing implementations, Robust PCA only takes one observation at a
    time into account (i.e., does not look at windows). However, Robust PCA can
    not be applied to a single variable, which is the case for univariate data.
    Therefore, we added a parameter ``window_size`` to apply Robust PCA in windows
    of a univariate time series, to make it applicable. Common behavior on multivariate
    time series can be obtained by setting ``window_size = 1``.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    max_iter: int, default=1000
        The maximum number of iterations allowed to optimize the low rank approximation.
    kwargs:
        Additional parameters to be passed PCA of Sklearn.

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pca_ : PCA
        The PCA-object used to project the data in a lower dimension.

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import RobustPrincipalComponentAnalysis
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> rpca = RobustPrincipalComponentAnalysis(2).fit(x)
    >>> rpca.decision_function(x)
    array([1.28436687, 1.29156655, 1.33793287, ..., 1.35563558, 1.25948662,
           1.2923824 ])

    References
    ----------
    .. [Candes2011robust] Emmanuel J. Candès, Xiaodong Li, Yi Ma, and John Wright. 2011. Robust principal
       component analysis? J. ACM 58, 3, Article 11 (June 2011), 37 pages. doi: `10.1145/1970392.1970395 <https://doi.org/10.1145/1970392.1970395>`_
    """

    window_size: Union[int, str]
    stride: int
    max_iter: int
    kwargs: dict
    window_size_: int
    pca_: PCA

    def __init__(
        self,
        window_size: Union[str, int],
        stride: int = 1,
        max_iter: int = 1000,
        **kwargs,
    ):
        super().__init__(Supervision.SEMI_SUPERVISED)

        check_is_valid_window_size(window_size)

        if not isinstance(max_iter, int) or isinstance(max_iter, bool):
            raise TypeError("`max_iter` should be an integer")
        if max_iter < 1:
            raise ValueError("`max_iter`should be at least 1")

        PCA(n_components=0.1, **kwargs)  # Check if PCA can be initialized

        self.window_size = window_size
        self.stride = stride
        self.max_iter = max_iter
        self.kwargs = kwargs

    def fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None, **kwargs
    ) -> "RobustPrincipalComponentAnalysis":
        """
        Fit this Robust PCA to the given data

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.
        y: ignored
            Not used, present for API consistency by convention.
        kwargs:
            Additional parameters to be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.

        Returns
        -------
        self: RobustPrincipleComponentAnalysis
            Returns the instance itself

        Raises
        ------
        ValueError
            If `X` is not a valid array.
        """
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")

        # Make sure X is a numpy array
        X = np.asarray(X)

        # Compute the windows
        self.window_size_ = compute_window_size(X, self.window_size, **kwargs)
        sliding_windows = sliding_window(X, self.window_size_, self.stride)

        # Apply robust PCA
        robust_pca = _RobustPCA(sliding_windows)
        L, S = robust_pca.fit(max_iter=self.max_iter)
        self.pca_ = PCA(n_components=L.shape[1], **self.kwargs)
        self.pca_.fit(L)
        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        if not hasattr(self, "pca_"):
            raise NotFittedError("Call the fit function before making predictions!")
        if not utils.is_valid_array_like(X):
            raise ValueError(f"Input must be numerical array-like")

        # Make sure X is a numpy array
        X = np.asarray(X)

        # Convert to sliding windows
        windows = sliding_window(X, self.window_size_, self.stride)

        # DO RPCA
        L = self.pca_.transform(windows)
        S = np.absolute(windows - L)
        per_window_decision_scores = S.sum(axis=1)

        # Get an anomaly score for each window
        decision_scores = reverse_sliding_window(
            per_window_decision_scores, self.window_size_, self.stride, X.shape[0]
        )

        return decision_scores


# From https://github.com/dganguli/robust-pca
class _RobustPCA:

    def __init__(self, D):
        self.D = D
        self.S = np.zeros(self.D.shape)
        self.Y = np.zeros(self.D.shape)

        self.mu = np.prod(self.D.shape) / (4 * np.linalg.norm(self.D, ord=1))
        self.mu_inv = 1 / self.mu
        self.lmbda = 1 / np.sqrt(np.max(self.D.shape))

    @staticmethod
    def frobenius_norm(M):
        return np.linalg.norm(M, ord="fro")

    @staticmethod
    def shrink(M, tau):
        return np.sign(M) * np.maximum((np.abs(M) - tau), np.zeros(M.shape))

    def svd_threshold(self, M, tau):
        U, S, V = np.linalg.svd(M, full_matrices=False)
        return np.dot(U, np.dot(np.diag(self.shrink(S, tau)), V))

    def fit(self, max_iter=1000):
        iter = 0
        err = np.inf
        Sk = self.S
        Yk = self.Y
        Lk = np.zeros(self.D.shape)

        _tol = 1e-7 * self.frobenius_norm(self.D)

        # this loop implements the principal component pursuit (PCP) algorithm
        # located in the table on page 29 of https://arxiv.org/pdf/0912.3599.pdf
        while (err > _tol) and iter < max_iter:
            Lk = self.svd_threshold(
                self.D - Sk + self.mu_inv * Yk, self.mu_inv
            )  # this line implements step 3
            Sk = self.shrink(
                self.D - Lk + (self.mu_inv * Yk), self.mu_inv * self.lmbda
            )  # this line implements step 4
            Yk = Yk + self.mu * (self.D - Lk - Sk)  # this line implements step 5
            err = self.frobenius_norm(self.D - Lk - Sk)
            iter += 1

        self.L = Lk
        self.S = Sk
        return Lk, Sk



import abc

import numpy as np

from dtaianomaly.evaluation.metrics import BinaryMetric
from dtaianomaly.evaluation.simple_binary_metrics import FBeta, Precision, Recall


def point_adjust(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """
    Apply point-adjustment to the given arrays. For each anomalous
    event in the ground truth (a sequence of consecutive anomalous
    observations), if any observation is predicted as an anomaly,
    all observations in the sequence are said to be detected.

    Parameters
    ----------
    y_true: array-like of shape (n_samples)
        Ground-truth labels.
    y_pred: array-like of shape (n_samples)
        Predicted anomaly scores.

    Returns
    -------
    point_adjusted_y_pred: array-like of shape (n_samples)
        The point adjusted predicted anomalies
    """
    # Find the anomalous events
    diff = np.diff(y_true, prepend=0, append=0)
    start_events = np.where(diff == 1)[0]
    end_events = np.where(diff == -1)[0]

    # Check if an anomaly is detected in any anomalous event
    point_adjusted_y_pred = y_pred.copy()
    for start, end in zip(start_events, end_events):
        if y_pred[start:end].any():
            point_adjusted_y_pred[start:end] = 1

    # Return the point adjusted scores
    return point_adjusted_y_pred


class PointAdjusted(BinaryMetric, abc.ABC):

    metric: BinaryMetric

    def __init__(self, metric: BinaryMetric):
        if not isinstance(metric, BinaryMetric):
            raise TypeError("The given `metric` should be a binary metric")
        self.metric = metric

    def _compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        point_adjusted_y_pred = point_adjust(y_true, y_pred)
        return self.metric._compute(y_true, point_adjusted_y_pred)


class PointAdjustedPrecision(PointAdjusted):
    """
    Compute the point-adjusted precision: first point-adjust the predicted
    anomaly scores, after which the precision is computed.

    For given binary anomaly predictions and ground truth anomaly labels,
    point-adjusting will treat any sequence of consecutive ground truth
    anomalies as anomalous events. If any of the observations in such an
    event has been detected, then we say that the anomaly has been detected.
    In this case, all predictions in the anomalous event are set to 1,
    thereby indicating that the method predicted an anomaly.

    Attributes
    ----------
    metric: Precision
        The Precision-object used for computing the precision, after the
        prediction has been point adjusted. Note that the object should
        not be passed to the constructor.

    See Also
    --------
    Precision: Compute the standard, not point-adjusted precision.
    """

    def __init__(self):
        super().__init__(Precision())


class PointAdjustedRecall(PointAdjusted):
    """
    Compute the point-adjusted recall: first point-adjust the predicted
    anomaly scores, after which the recall is computed.

    For given binary anomaly predictions and ground truth anomaly labels,
    point-adjusting will treat any sequence of consecutive ground truth
    anomalies as anomalous events. If any of the observations in such an
    event has been detected, then we say that the anomaly has been detected.
    In this case, all predictions in the anomalous event are set to 1,
    thereby indicating that the method predicted an anomaly.

    Attributes
    ----------
    metric: Recall
        The Recall-object used for computing the precision, after the
        recall has been point adjusted. Note that the object should
        not be passed to the constructor.

    See Also
    --------
    Recall: Compute the standard, not point-adjusted recall.
    """

    def __init__(self):
        super().__init__(Recall())


class PointAdjustedFBeta(PointAdjusted):
    """
    Compute the point-adjusted :math:`F_\\beta`: first point-adjust the predicted
    anomaly scores, after which the :math:`F_\\beta` is computed.

    For given binary anomaly predictions and ground truth anomaly labels,
    point-adjusting will treat any sequence of consecutive ground truth
    anomalies as anomalous events. If any of the observations in such an
    event has been detected, then we say that the anomaly has been detected.
    In this case, all predictions in the anomalous event are set to 1,
    thereby indicating that the method predicted an anomaly.

    Parameters
    ----------
    beta: int, float, default=1
        Desired beta parameter.

    Attributes
    ----------
    metric: FBeta
        The FBeta-object used for computing the precision, after the
        :math:`F_\\beta`: has been point adjusted. Note that the object should
        not be passed to the constructor.

    See Also
    --------
    FBeta: Compute the standard, not point-adjusted :math:`F_\\beta`.
    """

    def __init__(self, beta: (float, int) = 1):
        super().__init__(FBeta(beta))

    @property
    def beta(self):
        return self.metric.beta



from pyod.models.lof import LOF

from dtaianomaly.anomaly_detection.BaseDetector import Supervision
from dtaianomaly.anomaly_detection.PyODAnomalyDetector import PyODAnomalyDetector


class LocalOutlierFactor(PyODAnomalyDetector):
    """
    Anomaly detector based on the Local Outlier Factor.

    The local outlier factor [Breunig2000LOF]_ compares the density of each
    sample to the density of the neighboring samples. If the neighbors of a
    sample have a much higher density that the sample itself, the sample is
    considered anomalous. By looking at the local density (i.e., only comparing
    with the neighbors of a sample), the local outlier factor takes into
    account varying densities across the sample space.

    Notes
    -----
    The Local Outlier Factor inherets from :py:class:`~dtaianomaly.anomaly_detection.PyodAnomalyDetector`.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs:
        Arguments to be passed to the PyOD local outlier factor

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : LOF
        A LOF-detector of PyOD

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import LocalOutlierFactor
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> local_outlier_factor = LocalOutlierFactor(10).fit(x)
    >>> local_outlier_factor.decision_function(x)
    array([0.98370943, 0.98533454, 0.98738196, ..., 1.02394282, 1.02648068,
           1.01827158])

    References
    ----------
    .. [Breunig2000LOF] Markus M. Breunig, Hans-Peter Kriegel, Raymond T. Ng, and Jörg Sander.
       2000. LOF: identifying density-based local outliers. In Proceedings of the 2000 ACM
       SIGMOD international conference on Management of data (SIGMOD '00). Association for
       Computing Machinery, New York, NY, USA, 93–104. doi: `10.1145/342009.335388 <https://doi.org/10.1145/342009.335388>`_
    """

    def _initialize_detector(self, **kwargs) -> LOF:
        return LOF(**kwargs)

    def _supervision(self):
        return Supervision.UNSUPERVISED



from typing import Optional, Tuple

import numpy as np
from sklearn.exceptions import NotFittedError

from dtaianomaly.preprocessing.Preprocessor import Preprocessor
from dtaianomaly.utils import get_dimension


class RobustScaler(Preprocessor):
    """
    Scale the time series using robust statistics.

    The :py:class:`~dtaianomaly.preprocessing.RobustScaler` is similar to
    :py:class:`~dtaianomaly.preprocessing.StandardScaler`, but uses robust
    statistics rather than mean and standard deviation. The center of the data
    is computed via the median, and the scale is computed as the range between
    two quantiles (by default uses the IQR). This ensures that scaling is less
    affected by outliers.

    For a time series :math:`x`, center :math:`c` and scale :math:`s`, observation
    :math:`x_i` is scaled to observation :math:`y_i` using the following equation:

    .. math::

       y_i = \\frac{x_i - c}{s}

    Notice the similarity with the formula for standard scaling. For multivariate
    time series, each attribute is scaled independently, each with an independent
    scale and center.

    Parameters
    ----------
    quantile_range: tuple of (float, float), default = (25.0, 75.0)
        Quantile range used to compute the ``scale_`` of the robust scaler.
        By default, this is equal to the Inter Quantile Range (IQR). The first
        value of the quantile range corresponds to the smallest quantile, the
        second value corresponds to the larger quantile. If the first value is
        not smaller than the second value, an error will be thrown. The values
        must also both be in the range [0, 100].

    Attributes
    ----------
    center_: array-like of shape (n_attributes)
        The median value in each attribute of the training data.
    scale_: array-like of shape (n_attributes)
        The quantile range for each attribute of the training data.

    Raises
    ------
    NotFittedError
        If the `transform` method is called before fitting this StandardScaler.
    """

    quantile_range: (float, float)
    center_: np.array
    scale_: np.array

    def __init__(self, quantile_range: (float, float) = (25.0, 75.0)):
        if not isinstance(quantile_range, tuple):
            raise TypeError("`quantile_range` should be tuple")
        if len(quantile_range) != 2:
            raise ValueError(
                "'quantile_range' should consist of exactly two values (length of 2)"
            )
        if not isinstance(quantile_range[0], (float, int)) or isinstance(
            quantile_range[0], bool
        ):
            raise TypeError(
                "The first element `quantile_range` should be a float or int"
            )
        if not isinstance(quantile_range[1], (float, int)) or isinstance(
            quantile_range[1], bool
        ):
            raise TypeError(
                "The second element `quantile_range` should be a float or int"
            )
        if quantile_range[0] < 0.0:
            raise ValueError(
                "the first element in 'quantile_range' must be at least 0.0"
            )
        if quantile_range[1] > 100.0:
            raise ValueError(
                "the second element in 'quantile_range' must be at most 100.0"
            )
        if not quantile_range[0] < quantile_range[1]:
            raise ValueError(
                "the first element in 'quantile_range' must be at smaller than the second element in 'quantile_range'"
            )
        self.quantile_range = quantile_range

    def _fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "RobustScaler":
        if get_dimension(X) == 1:
            # univariate case
            self.center_ = np.array([np.nanmedian(X)])
            q_min = np.percentile(X, q=self.quantile_range[0])
            q_max = np.percentile(X, q=self.quantile_range[1])
            self.scale_ = np.array([q_max - q_min])
        else:
            # multivariate case
            self.center_ = np.nanmedian(X, axis=0)
            q_min = np.percentile(X, q=self.quantile_range[0], axis=0)
            q_max = np.percentile(X, q=self.quantile_range[1], axis=0)
            self.scale_ = q_max - q_min
        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        if not (hasattr(self, "center_") and hasattr(self, "scale_")):
            raise NotFittedError(f"Call `fit` before using transform on {str(self)}")
        if not (
            (len(X.shape) == 1 and self.center_.shape[0] == 1)
            or X.shape[1] == self.center_.shape[0]
        ):
            raise AttributeError(
                f"Trying to robust scale a time series with {X.shape[0]} attributes while it was fitted on {self.center_.shape[0]} attributes!"
            )

        X_ = (X - self.center_) / self.scale_
        return np.where(np.isnan(X_), X, X_), y



import abc

import numpy as np

from dtaianomaly import utils
from dtaianomaly.PrettyPrintable import PrettyPrintable
from dtaianomaly.thresholding import Thresholding


class Metric(PrettyPrintable):

    def compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """
        Computes the performance score.

        Parameters
        ----------
        y_true: array-like of shape (n_samples)
            Ground-truth labels.
        y_pred: array-like of shape (n_samples)
            Predicted anomaly scores.

        Returns
        -------
        score: float
            The alignment score of the given ground truth and
            prediction, according to this score.

        Raises
        ------
        ValueError
            When inputs are not numeric "array-like"s
        ValueError
            If shapes of `y_true` and `y_pred` are not of identical shape
        ValueError
            If `y_true` is non-binary.
        """
        if not utils.is_valid_array_like(y_true):
            raise ValueError("Input 'y_true' should be numeric array-like")
        if not utils.is_valid_array_like(y_pred):
            raise ValueError("Input 'y_pred' should be numeric array-like")
        y_true = np.asarray(y_true)
        y_pred = np.asarray(y_pred)
        if not y_true.shape == y_pred.shape:
            raise ValueError("Inputs should have identical shape")
        if not np.all(np.isin(y_true, [0, 1])):
            raise ValueError("The predicted anomaly scores must be binary!")
        return self._compute(y_true, y_pred)

    @abc.abstractmethod
    def _compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """Effectively compute the metric."""


class ProbaMetric(Metric, abc.ABC):
    """A metric that takes as input continuous anomaly scores."""


class BinaryMetric(Metric, abc.ABC):
    """A metric that takes as input binary anomaly labels."""

    def compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """
        Computes the performance score.

        Parameters
        ----------
        y_true: array-like of shape (n_samples)
            Ground-truth labels.
        y_pred: array-like of shape (n_samples)
            Predicted anomaly scores.

        Returns
        -------
        score: float
            The alignment score of the given ground truth and
            prediction, according to this score.

        Raises
        ------
        ValueError
            When inputs are not numeric "array-like"s
        ValueError
            If shapes of `y_true` and `y_pred` are not of identical shape
        ValueError
            If `y_true` is non-binary.
        ValueError
            If `y_pred` is non-binary.
        """
        if not np.all(np.isin(y_pred, [0, 1])):
            raise ValueError("The predicted anomaly scores must be binary!")
        return super().compute(y_true, y_pred)


class ThresholdMetric(ProbaMetric):
    """
    Wrapper to combine a `BinaryMetric` object with some
    thresholding, to make sure that it can take continuous
    anomaly scores as an input. This is done by first applying
    some thresholding to the predicted anomaly scores, after
    which a binary metric can be computed.

    Parameters
    ----------
    thresholder: Thresholding
        Instance of the desired `Thresholding` class
    metric: Metric
        Instance of the desired `Metric` class
    """

    thresholder: Thresholding
    metric: BinaryMetric

    def __init__(self, thresholder: Thresholding, metric: BinaryMetric) -> None:
        if not isinstance(thresholder, Thresholding):
            raise TypeError(
                f"thresholder expects 'Thresholding', got {type(thresholder)}"
            )
        if not isinstance(metric, BinaryMetric):
            raise TypeError(f"metric expects 'BinaryMetric', got {type(metric)}")
        super().__init__()
        self.thresholder = thresholder
        self.metric = metric

    def _compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        y_pred_binary = self.thresholder.threshold(y_pred)
        # Can compute the inner method, because checks have already been done at this point.
        return self.metric._compute(y_true=y_true, y_pred=y_pred_binary)

    def __str__(self) -> str:
        return f"{self.thresholder}->{self.metric}"



from pyod.models.knn import KNN

from dtaianomaly.anomaly_detection.BaseDetector import Supervision
from dtaianomaly.anomaly_detection.PyODAnomalyDetector import PyODAnomalyDetector


class KNearestNeighbors(PyODAnomalyDetector):
    """
    Anomaly detector based on K-nearest neighbors [ramaswamy2000efficient]_.

    Given some distance metric :math:`dist`, the :math:`K`-nearest neighbor of an
    instance :math:`x` is the sample :math:`y` such that there exist exactly :math:`K-1`
    other samples :math:`z` with :math:`dist(x, z) < dist(x, y)`. The :math:`K`-nearest neighbor
    distance of :math:`x` equals the distance to this :math:`K`the nearest neighbor.
    The larger this :math:`K`-nearest neighbor distance of a sample is, the further
    away it is from the other instances. :math:`K`-nearest neighbor uses this distance
    as an anomaly score, and thus detects distance-based anomalies.

    Notes
    -----
    The K-nearest neighbors inherets from :py:class:`~dtaianomaly.anomaly_detection.PyodAnomalyDetector`.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs
        Arguments to be passed to the PyOD isolation forest.

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : KNN
        A K-nearest neighbors detector of PyOD

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import KNearestNeighbors
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> knn = KNearestNeighbors(10).fit(x)
    >>> knn.decision_function(x)
    array([0.2527578 , 0.26430228, 0.2728953 , ..., 0.26269151, 0.26798469,
           0.26139759])

    References
    ----------
    .. [ramaswamy2000efficient] Ramaswamy, Sridhar, Rajeev Rastogi, and Kyuseok Shim.
       "Efficient algorithms for mining outliers from large data sets." Proceedings
       of the 2000 ACM SIGMOD international conference on Management of data. 2000,
       doi: `10.1145/342009.33543 <https://doi.org/10.1145/342009.33543>`_.
    """

    def _initialize_detector(self, **kwargs) -> KNN:
        return KNN(**kwargs)

    def _supervision(self):
        return Supervision.UNSUPERVISED



import numpy as np
from sklearn import metrics

from dtaianomaly.evaluation.metrics import ProbaMetric


class AreaUnderROC(ProbaMetric):
    """
    Computes the Area Under the Receiver Operating Characteristic Curve (AUC-ROC) score.

    The AUC-ROC is a widely used metric to evaluate the performance of
    a binary classifier, especially in anomaly detection. The ROC-curve
    plots the true positive rate (recall) against the false positive
    rate across different classification thresholds. The AUC-ROC represents
    the likelihood that the model ranks a randomly chosen anomaly higher
    than a randomly chosen normal instance.

    AUC-ROC provides a single number summarizing the model's ability to
    distinguish between normal and anomalous instances. A value of 1.0
    indicates perfect discrimination, while 0.5 implies the model performs
    no better than random guessing. It is especially useful when anomalies
    are rare, as it considers the trade-off between detecting true anomalies
    (high recall) and minimizing false positives.
    """

    def __init__(self) -> None:
        super().__init__()

    def _compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        return float(metrics.roc_auc_score(y_true=y_true, y_score=y_pred))


class AreaUnderPR(ProbaMetric):
    """
    Computes the Area Under the Precision-Recall Curve (AUC-PR) score.

    The AUC-PR is a performance metric that is especially useful for
    evaluating models in imbalanced datasets, such as anomaly detection,
    where the number of normal instances vastly outnumbers the anomalies.
    The Precision-Recall curve plots precision against recall at various
    thresholds, providing a detailed view of the trade-off between
    detecting true anomalies (recall) and minimizing false alarms (precision).

    AUC-PR summarizes the curve into a single value, representing the overall
    ability of the model to identify anomalies while keeping false positives
    in check. A higher AUC-PR value indicates better performance, meaning the
    model is effective at detecting true anomalies with fewer false positives.
    """

    def __init__(self) -> None:
        super().__init__()

    def _compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        precision, recall, _ = metrics.precision_recall_curve(y_true, y_pred)
        return float(metrics.auc(recall, precision))



from typing import Optional

import numpy as np

from dtaianomaly import utils
from dtaianomaly.anomaly_detection.BaseDetector import BaseDetector, Supervision


class AlwaysNormal(BaseDetector):
    """
    Baseline anomaly detector, which predicts that all observations are normal.
    This detector should only be used for sanity-check, and not to effectively
    detect anomalies in time series data.
    """

    def __init__(self):
        super().__init__(Supervision.UNSUPERVISED)

    def fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None, **kwargs
    ) -> "AlwaysNormal":
        """
        Simply return this detector, because no fitting is required.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.
        y: ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self: AlwaysNormal
            Returns the instance itself
        """
        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        """
        Predicts 'always normal' anomaly scores, i.e., always returns a '0.0'.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.

        Returns
        -------
        normal_scores: array-like of shape (n_samples)
            All normal anomaly scores.

        Raises
        ------
        ValueError
            If `X` is not a valid array.
        """
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")
        return np.zeros(shape=np.asarray(X).shape[0])


class AlwaysAnomalous(BaseDetector):
    """
    Baseline anomaly detector, which predicts that all observations are anomalous.
    This detector should only be used for sanity-check, and not to effectively
    detect anomalies in time series data.
    """

    def __init__(self):
        super().__init__(Supervision.UNSUPERVISED)

    def fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None, **kwargs
    ) -> "AlwaysAnomalous":
        """
        Simply return this detector, because no fitting is required.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.
        y: ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self: AlwaysAnomalous
            Returns the instance itself
        """
        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        """
        Predicts 'always anomalous' anomaly scores, i.e., always returns a '1.0'.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.

        Returns
        -------
        anomalous_scores: array-like of shape (n_samples)
            All anomalous anomaly scores.

        Raises
        ------
        ValueError
            If `X` is not a valid array.
        """
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")
        return np.ones(shape=np.asarray(X).shape[0])


class RandomDetector(BaseDetector):
    """
    Baseline anomaly detector, which assigns random anomaly scores. This detector
    should only be used for sanity-check, and not to effectively detect anomalies
    in time series data.

    Parameters
    ----------
    seed: int, default=None
        The seed to use for generating anomaly scores. If None, no seed will be used.
    """

    seed: Optional[int]

    def __init__(self, seed: Optional[int] = None):
        super().__init__(Supervision.UNSUPERVISED)
        self.seed = seed

    def fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None, **kwargs
    ) -> "RandomDetector":
        """
        Simply return this detector, because no fitting is required.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.
        y: ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self: RandomDetector
            Returns the instance itself
        """
        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        """
        Predicts random anomaly scores. Uses numpy random-number generator, without
        adjusting the internal seed of numpy.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.

        Returns
        -------
        random_scores: array-like of shape (n_samples)
            Random anomaly scores.

        Raises
        ------
        ValueError
            If `X` is not a valid array.
        """
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")
        rng = np.random.default_rng(seed=self.seed)
        return rng.random(size=np.asarray(X).shape[0])



from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import ConnectionPatch

from dtaianomaly import utils


def plot_time_series_colored_by_score(
    X: np.ndarray,
    y: np.ndarray,
    time_steps: np.array = None,
    ax: plt.Axes = None,
    nb_colors: int = 100,
    **kwargs,
) -> plt.Figure:
    """
    Plots the given time series, and color it according to the given scores.
    Higher scores will be colored red, and lower scores will be colored green.
    Thus, if the ground truth anomaly scores are passed, red corresponds to
    anomalies and green to normal observations.

    Parameters
    ----------
    X: np.ndarray of shape (n_samples, n_attributes)
        The time series to plot
    y: np.ndarray of shape (n_samples)
        The scores, according to which the plotted data should be colored.
    time_steps: np.array of shape (n_samples), default=None
        The time steps to plot. If no time steps are provided, then the
        default range ``[0, ..., n_samples-1]`` will be used.
    ax: plt.Axes, default=None
        The axes onto which the plot should be made. If None, then a new
        figure and axis will be created.
    nb_colors: int, default=100
        The number of colors to use for plotting the time series.
    **kwargs:
        Arguments to be passed to plt.Figure(), in case ``ax=None``.

    Returns
    -------
    fig: plt.Figure
        The figure containing the plotted data.

    Notes
    -----
    Each segment in the time series will be plotted independently. Thus,
    for time series with many observations, plotting the data using this
    method can cost a huge amount of time.
    """
    if ax is None:
        plt.figure(**kwargs)
        ax = plt.gca()

    # Format the time steps
    time_steps = format_time_steps(time_steps, X.shape[0])

    y_min, y_max = y.min(), y.max()
    y_scaled = (y - y_min) / (y_max - y_min) if y_max > y_min else np.zeros_like(y)
    y_binned = [np.floor(score * nb_colors) / nb_colors for score in y_scaled]
    colormap = plt.get_cmap("RdYlGn", nb_colors).reversed()
    for i in range(0, X.shape[0] - 1):
        color = colormap(y_binned[i])
        ax.plot([time_steps[i], time_steps[i + 1]], X[[i, i + 1]], c=color)
    return plt.gcf()


def plot_time_series_anomalies(
    X: np.ndarray,
    y: np.ndarray,
    y_pred: np.ndarray,
    time_steps: np.array = None,
    ax: plt.Axes = None,
    **kwargs,
) -> plt.Figure:
    """
    Visualizes time series data with true and predicted anomalies, highlighting true positives (TP),
    false positives (FP), and false negatives (FN).

    Parameters
    ----------
    X: np.ndarray of shape (n_samples, n_attributes)
        The time series to plot
    y: np.ndarray of shape (n_samples,)
        Ground truth anomaly labels (binary values: 0 or 1).
    y_pred: np.ndarray of shape (n_samples,)
        Predicted anomaly labels (binary values: 0 or 1).
    time_steps: np.array of shape (n_samples), default=None
        The time steps to plot. If no time steps are provided, then the
        default range ``[0, ..., n_samples-1]`` will be used.
    ax: plt.Axes, default=None
        The axes onto which the plot should be made. If None, then a new
        figure and axis will be created.
    **kwargs:
        Arguments to be passed to plt.Figure(), in case ``ax=None``.

    Returns
    -------
    fig: plt.Figure
        The figure containing the plotted data.
    """

    # Prepare the axis
    if ax is None:
        plt.figure(**kwargs)
        ax = plt.gca()

    # Check if the given y values are binary
    if not np.all(np.isin(y, [0, 1])):
        raise ValueError("The predicted anomaly scores must be binary.")
    if not np.all(np.isin(y_pred, [0, 1])):
        raise ValueError("The predicted anomaly scores must be binary.")

    # Identify TP, FP, FN
    TP = (y == 1) & (y_pred == 1)
    FP = (y == 0) & (y_pred == 1)
    FN = (y == 1) & (y_pred == 0)

    # Format the time steps
    time_steps = format_time_steps(time_steps, X.shape[0])

    # Plot the time series
    ax.plot(time_steps, X)

    # Scatter points for TP, FP, FN
    X_reshaped = X.reshape((-1, utils.get_dimension(X)))
    tps, fps, fns = None, None, None
    for i in range(utils.get_dimension(X)):
        tps = ax.scatter(time_steps[TP], X_reshaped[TP, i], color="green")
        fps = ax.scatter(time_steps[FP], X_reshaped[FP, i], color="red")
        fns = ax.scatter(time_steps[FN], X_reshaped[FN, i], color="orange")
    ax.legend([tps, fps, fns], ["TP", "FP", "FN"])

    return plt.gcf()


def plot_demarcated_anomalies(
    X: np.ndarray,
    y: np.array,
    ax: plt.Axes = None,
    time_steps: np.array = None,
    color_anomaly: str = "red",
    alpha_anomaly: float = 0.2,
    **kwargs,
) -> plt.Figure:
    """
    Plot the given time series and binary anomaly labels. Each anomalous
    interval is marked by a colored area, depending on the provided parameters.

    Parameters
    ----------
    X: np.ndarray of shape (n_samples, n_attributes)
        The time series to plot
    y: np.array of shape (n_samples)
        The binary anomaly scores.
    ax: plt.Axes, default=None
        The axes onto which the plot should be made. If None, then a new
        figure and axis will be created.
    time_steps: np.array of shape (n_samples), default=None
        The time steps to plot. If no time steps are provided, then the
        default range ``[0, ..., n_samples-1]`` will be used.
    color_anomaly: str, default='red'
        The color in which the anomaly should be marked.
    alpha_anomaly: float, default=0.2
        The alpha value for marking the anomaly, to adjust transparency.
    **kwargs:
        Arguments to be passed to plt.Figure(), in case ``ax=None``.

    Returns
    -------
    fig: plt.Figure
        The figure containing the plotted data.
    """
    # Check if y is binary
    if not np.all(np.isin(y, [0, 1])):
        raise ValueError("The predicted anomaly scores must be binary!")

    # Initialize an axis object if none has been given
    if ax is None:
        plt.figure(**kwargs)
        ax = plt.gca()

    # Identify the anomalous regions
    diff = np.diff(y, prepend=0, append=0)
    start_events = np.where(diff == 1)[0]
    end_events = np.where(diff == -1)[0]

    # Format the time steps
    time_steps = format_time_steps(time_steps, X.shape[0])

    # Plot the time series data
    ax.plot(time_steps, X)

    # Plot the anomalous zones
    for start, end in zip(start_events, end_events):
        ax.axvspan(
            time_steps[start],
            time_steps[min(end, time_steps.shape[0] - 1)],
            color=color_anomaly,
            alpha=alpha_anomaly,
        )

    # Return the active figure
    return plt.gcf()


def plot_with_zoom(
    X: np.ndarray,
    y: np.array,
    start_zoom: int,
    end_zoom: int,
    time_steps: np.array = None,
    y_pred: np.array = None,
    method_to_plot=plot_demarcated_anomalies,
    color: str = "blue",
    linewidth: float = 3,
    linestyle: str = "--",
    **kwargs,
) -> plt.Figure:
    """
    Plot the given data in two axes, one showing the entire time
    series and one zooming in on a specific area of the time series.

    Parameters
    ----------
    X: np.ndarray of shape (n_samples, n_attributes)
        The time series to plot
    y: np.array of shape (n_samples)
        The binary anomaly scores.
    start_zoom: int
        The index in the data at which the zoom starts.
    end_zoom: int
        The index in the data at which the zoom ends.
    time_steps: np.array of shape (n_samples), default=None
        The time steps to plot. If no time steps are provided, then the
        default range ``[0, ..., n_samples-1]`` will be used.
    y_pred: np.array of shape (n_samples), default=None
        The predicted anomaly scores to plot. Is necessary if the
        ``method_to_plot`` requires predicted anomaly scores.
    method_to_plot: callable, default=:py:autofunc:`~dtaianomaly.visualization.plot_demarcated_anomalies`
        Method used for plotting the data. Should take as inputs
        the values ``X`` (the time series data), ``y`` (the anomaly
        labels``), ``time_steps`` (the time steps at which there was
        an observation) and ``ax`` (the axis on which the plot should
        be made). Optionally, the method takes as input a value ``y_pred``
        for the predicted anomaly scores.
    color: string, default='blue'
        The color of the lines to demarcate the area of zooming.
    linewidth: float, default=3
        The width of the lines to demarcate the area of zooming.
    linestyle: str, default='--'
        The style of the lines to demarcate the area of zooming.
    **kwargs:
        Arguments to be passed to plt.subplots().

    Returns
    -------
    fig: plt.Figure
        The figure containing the plotted data.
    """
    # Create the main figure and two subplots (axes)
    fig, (ax_main, ax_zoom) = plt.subplots(2, 1, **kwargs)

    # Format the time steps
    time_steps = format_time_steps(time_steps, X.shape[0])

    # Plot the data
    X_zoom = X[start_zoom:end_zoom]
    y_zoom = y[start_zoom:end_zoom]
    time_stamps_zoom = time_steps[start_zoom:end_zoom]
    if y_pred is None:
        method_to_plot(X=X, y=y, ax=ax_main, time_steps=time_steps)
        method_to_plot(X=X_zoom, y=y_zoom, ax=ax_zoom, time_steps=time_stamps_zoom)
    else:
        method_to_plot(X=X, y=y, y_pred=y_pred, ax=ax_main, time_steps=time_steps)
        method_to_plot(
            X=X_zoom,
            y=y_zoom,
            y_pred=y_pred[start_zoom:end_zoom],
            ax=ax_zoom,
            time_steps=time_stamps_zoom,
        )

    # Draw vertical lines to demarcate the area in which is zoomed
    for ax in [ax_main, ax_zoom]:
        for x in [start_zoom, end_zoom]:
            ax.axvline(
                x=time_steps[x], color=color, linestyle=linestyle, linewidth=linewidth
            )

    # Connect the demarcations across the subplots
    fig.add_artist(
        ConnectionPatch(
            xyA=(time_steps[start_zoom], ax_main.get_ylim()[0]),
            coordsA=ax_main.transData,
            xyB=(time_steps[start_zoom], ax_zoom.get_ylim()[1]),
            coordsB=ax_zoom.transData,
            color=color,
            linestyle=linestyle,
            linewidth=linewidth,
        )
    )
    fig.add_artist(
        ConnectionPatch(
            xyA=(time_steps[end_zoom], ax_main.get_ylim()[0]),
            coordsA=ax_main.transData,
            xyB=(time_steps[end_zoom], ax_zoom.get_ylim()[1]),
            coordsB=ax_zoom.transData,
            color=color,
            linestyle=linestyle,
            linewidth=linewidth,
        )
    )

    return fig


def plot_anomaly_scores(
    X: np.array,
    y: np.array,
    y_pred: np.array,
    time_steps: np.array = None,
    method_to_plot=plot_demarcated_anomalies,
    confidence: np.array = None,
    **kwargs,
) -> plt.Figure:
    """
    Plot the given data with the ground truth anomalies, and compare the
    predicted anomaly scores.

    Parameters
    ----------
    X: np.ndarray of shape (n_samples, n_attributes)
        The time series to plot
    y: np.array of shape (n_samples)
        The binary anomaly scores.
    y_pred: np.array of shape (n_samples)
        The predicted anomaly scores to plot.
    time_steps: np.array of shape (n_samples), default=None
        The time steps to plot. If no time steps are provided, then the
        default range ``[0, ..., n_samples-1]`` will be used.
    method_to_plot: callable, default=:py:autofunc:`~dtaianomaly.visualization.plot_demarcated_anomalies`
        Method used for plotting the data along with the ground truth
        anomaly scores. Should take as inputs the values ``X`` (the
        time series data), ``y`` (the anomaly labels``), ``time_steps``
        (the time steps at which there was an observation) and ``ax``
        (the axis on which the plot should be made).
    confidence: np.array of shape (n_samples), default=None
        The confidence of the anomaly scores.
    **kwargs:
        Arguments to be passed to plt.subplots().

    Returns
    -------
    fig: plt.Figure
        The figure containing the plotted data.
    """
    # Create the figure
    fig, (ax_data, ax_pred) = plt.subplots(nrows=2, ncols=1, sharex=True, **kwargs)

    # Format the time steps
    time_steps = format_time_steps(time_steps, X.shape[0])

    # Plot the time series data
    ax_data.set_title("Time series data")
    method_to_plot(X=X, y=y, ax=ax_data, time_steps=time_steps)

    # Plot the anomaly scores
    ax_pred.set_title("Predicted anomaly scores")
    ax_pred.plot(time_steps, y_pred, label="Anomaly scores")

    # Predict the confidence interval
    if confidence is not None:
        ax_pred.fill_between(
            time_steps,
            y_pred - (1 - confidence),
            y_pred + (1 - confidence),
            color="gray",
            alpha=0.5,
            label="Confidence range",
        )
        ax_pred.legend()

    # Return the figure
    return fig


def format_time_steps(time_steps: Optional[np.array], n_samples: int) -> np.array:
    """
    Format the given time steps, to ensure that fixed time steps are
    provided in case they are ``None``. These fixed time steps will
    equal the range ``[0, ..., n_samples-1]``

    Parameters
    ----------
    time_steps: np.array of shape (n_samples) or None
        The time seps to format.
    n_samples: int
        The number of samples for which there should be a time step.

    Returns
    -------
    formatted_time_steps: np.array of shape (n_samples)
        If the given ``time_steps`` did not equal ``None``, then these
        values are returned. Otherwise, an array with values ``[0, ...,
        n_samples-1]`` is returned.
    """
    return np.arange(n_samples) if time_steps is None else time_steps



from pyod.models.cblof import CBLOF

from dtaianomaly.anomaly_detection.BaseDetector import Supervision
from dtaianomaly.anomaly_detection.PyODAnomalyDetector import PyODAnomalyDetector


class ClusterBasedLocalOutlierFactor(PyODAnomalyDetector):
    """
    Anomaly detector based on the Cluster-based Local Outlier Factor (CBLOF).

    CBLOF [he2003discovering]_ is a cluster-based LOF which uses the distance to
    clusters in the data to compute an outlier score. Specifically, CBLOF first
    clusters the data using some clustering algorithm (by default K-means). Next,
    the clusters are separated in the so-called 'large clusters' $LC$ and 'small
    clusters' $SC$, depending on the parameters :math:`\\alpha` and :math:`\\beta`.
    Then, the Cluster-based Local outlier Factor of an observation :math:`o` belonging
    to cluster :math:`C_i` is computed as follows:

    .. math::

       \\begin{equation}
           CBLOF(o) = \\lvert C_i \\rvert \\cdot
           \\begin{cases}
               dist(o, C_i), & \\text{if $C_i \\in LC$}. \\\\
               min_{C_j \\in LC} (dist(o, C_j)), & \\text{if $C_i \\in SC$}.
           \\end{cases}
       \\end{equation}

    Specifically, if :math:`o` is part of a large cluster :math:`C_i`, we multiply the size
    of :math:`C_i` with the distance of :math:`o` to  :math:`C_i`. If :math:`o` is in a small
    cluster, then the size of  :math:`C_i` is multiplied by the distance to the nearest
    *large* cluster  :math:`C_j`.

    Notes
    -----
    CBLOF inherets from :py:class:`~dtaianomaly.anomaly_detection.PyodAnomalyDetector`.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs
        Arguments to be passed to the PyOD CBLOF.

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : CBLOF
        A CBLOF detector of PyOD

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import ClusterBasedLocalOutlierFactor
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> cblof = ClusterBasedLocalOutlierFactor(10).fit(x)
    >>> cblof.decision_function(x)
    array([0.50321076, 0.5753145 , 0.61938076, ..., 0.29794485, 0.30720306,
           0.29857479])

    References
    ----------
    .. [he2003discovering] He, Zengyou, Xiaofei Xu, and Shengchun Deng. "Discovering
       cluster-based local outliers." Pattern recognition letters 24.9-10 (2003): 1641-1650,
       doi: `10.1016/S0167-8655(03)00003-5 <https://doi.org/10.1016/S0167-8655(03)00003-5>`_.
    """

    def _initialize_detector(self, **kwargs) -> CBLOF:
        return CBLOF(**kwargs)

    def _supervision(self):
        return Supervision.UNSUPERVISED



from typing import Optional, Tuple

import numpy as np

from dtaianomaly.preprocessing.Preprocessor import Preprocessor


class SamplingRateUnderSampler(Preprocessor):
    """
    Undersample time series with sampling rate `sampling_rate`. This means
    that every `sampling_rate` element is taken from the time series. After
    undersampling, only `1/sampling_rate` percent of the original samples
    will remain.

    Parameters
    ----------
    sampling_rate: int
        The rate at which should be sampled.
    """

    sampling_rate: int

    def __init__(self, sampling_rate: int) -> None:
        if sampling_rate <= 0:
            raise ValueError("Sampling rate should be strictly positive.")
        self.sampling_rate = sampling_rate

    def _fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> "SamplingRateUnderSampler":
        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        if self.sampling_rate >= X.shape[0]:
            raise ValueError(
                f"The sampling rate ('{self.sampling_rate}') is too large for a time series of shape {X.shape}!"
            )
        return X[:: self.sampling_rate], (
            None if y is None else y[:: self.sampling_rate]
        )


class NbSamplesUnderSampler(Preprocessor):
    """
    Undersample time series such that exactly `nb_samples` samples
    remain in the original time series. This enables to manually
    set the size of the transformed time series, independent of
    the original size of the time series.

    Parameters
    ----------
    nb_samples: int, default=None
        The number of samples remaining.
    """

    nb_samples: int

    def __init__(self, nb_samples: int) -> None:
        if nb_samples <= 1:
            raise ValueError("Number of samples should be at least 2.")
        self.nb_samples = nb_samples

    def _fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> "NbSamplesUnderSampler":
        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        if self.nb_samples >= X.shape[0]:
            return X, y
        indices = np.linspace(
            0, X.shape[0] - 1, self.nb_samples, dtype=int, endpoint=True
        )
        return X[indices], (None if y is None else y[indices])



import math
from typing import Union

import numpy as np
import scipy
from statsmodels.tsa.stattools import acf

from dtaianomaly import utils


def sliding_window(X: np.ndarray, window_size: int, stride: int) -> np.ndarray:
    """
    Constructs a sliding window for the given time series.

    Parameters
    ----------
    X: array-like of shape (n_samples, n_attributes)
        The time series
    window_size: int
        The window size for the sliding windows.
    stride: int
        The stride, i.e., the step size for the windows.

    Returns
    -------
    windows: np.ndarray of shape ((n_samples - window_size)/stride + 1, n_attributes * window_size)
        The windows as a 2D numpy array. Each row corresponds to a
        window. For windows of multivariate time series are flattened
        to form a 1D array of length the number of attributes multiplied
        by the window size.
    """
    windows = [
        X[t : t + window_size].ravel()
        for t in range(0, X.shape[0] - window_size, stride)
    ]
    windows.append(X[-window_size:].ravel())
    return np.array(windows)


def reverse_sliding_window(
    per_window_anomaly_scores: np.ndarray,
    window_size: int,
    stride: int,
    length_time_series: int,
) -> np.ndarray:
    """
    Reverses the sliding window, to convert the per-window anomaly
    scores into per-observation anomaly scores.

    For non-overlapping sliding windows, it is trivial to convert
    the per-window anomaly scores to per-observation scores, because
    each observation is linked to only one window. For overlapping
    windows, certain observations are linked to one or more windows
    (depending on the window size and stride), obstructing simply
    copying the corresponding per-window anomaly score to each window.
    In the case of multiple overlapping windows, the anomaly score
    of the observation is set to the mean of the corresponding
    per-window anomaly scores.

    Parameters
    ----------
    per_window_anomaly_scores: array-like of shape (n_windows)
    window_size: int
        The window size used for creating windows
    stride: int
        The stride, i.e., the step size used for creating windows
    length_time_series: int
        The original length of the time series.

    Returns
    -------
    anomaly_scores: np.ndarray of shape (length_time_series)
        The per-observation anomaly scores.
    """
    # Convert to array
    scores_time = np.empty(length_time_series)

    start_window_index = 0
    min_start_window = 0
    end_window_index = 0
    min_end_window = 0
    for t in range(length_time_series - window_size):
        while min_start_window + window_size <= t:
            start_window_index += 1
            min_start_window += stride
        while t >= min_end_window:
            end_window_index += 1
            min_end_window += stride
        scores_time[t] = np.mean(
            per_window_anomaly_scores[start_window_index:end_window_index]
        )

    for t in range(length_time_series - window_size, length_time_series):
        while min_start_window + window_size <= t:
            start_window_index += 1
            min_start_window += stride
        scores_time[t] = np.mean(per_window_anomaly_scores[start_window_index:])

    return scores_time


def check_is_valid_window_size(window_size: Union[int, str]) -> None:
    """
    Checks if the given window size is valid or not. If the window size
    is not valid, a ValueError will be raised. Valid window sizes include:

    - a strictly positive integer
    - a string from the set {``'fft'``, ``'acf'``, ``'mwf'``, ``'suss'``}

    Parameters
    ----------
    window_size: int or string
        The valid to check if it is valid or not.

    Raises
    ------
    ValueError
        If the given ``window_size`` is not a valid window size.
    """
    if isinstance(window_size, int):
        if isinstance(window_size, bool):
            raise ValueError("The window size can not be a boolean value!")
        if window_size <= 0:
            raise ValueError("An integer window size should be strictly positive.")

    elif window_size not in ["fft", "acf", "mwf", "suss"]:
        raise ValueError(f"Invalid window_size given: '{window_size}'.")


def compute_window_size(
    X: np.ndarray,
    window_size: Union[int, str],
    lower_bound: int = 10,
    upper_bound: int = 1000,
    threshold: float = 0.89,
    default_window_size: int = None,
) -> int:
    """
    Compute the window size of the given time series [ermshaus2023window]_.

    Parameters
    ----------
    X: array-like of shape (n_samples, n_attributes)
        Input time series.

    window_size: int or str
        The method by which a window size should be computed. Valid options are:

        - ``int``: Simply return the given window size.
        - ``'fft'``: Compute the window size by selecting the dominant Fourier frequency.
        - ``'acf'``: Compute the window size as the leg with the highest autocorrelation.
        - ``'mwf'``: Computes the window size using the Multi-Window-Finder method [shima2021multi]_.
        - ``'suss'``: Computes the window size using the Summary Statistics Subsequence method [ermshaus2023clasp]_.

    lower_bound: int, default=10
        The lower bound on the automatically computed window size. Only used if ``window_size``
        equals ``'fft'``, ``'acf'``, ``'mwf'`` or ``'suss'``.

    upper_bound: int, default=1000
        The lower bound on the automatically computed window size. Only used if ``window_size``
        equals ``'fft'``, ``'acf'``, or ``'mwf'``.

    threshold: float, default=0.89
        The threshold for selecting the optimal window size using ``'suss'``.

    default_window_size: int, default=None
        The default window size, in case an invalid automatic window size was computed.
        By default, the value is set to None, which means that an error is thrown.

    Returns
    -------
    window_size_: int
        The computed window size.

    References
    ----------
    .. [ermshaus2023window] Ermshaus, Arik, Patrick Schäfer, and Ulf Leser. "Window
       size selection in unsupervised time series analytics: A review and benchmark."
       International Workshop on Advanced Analytics and Learning on Temporal Data.
       Springer, Cham, 2023, doi: `10.1007/978-3-031-24378-3_6 <https://doi.org/10.1007/978-3-031-24378-3_6>`_

    .. [shima2021multi] Imani, Shima, and Eamonn Keogh. "Multi-window-finder: domain
       agnostic window size for time series data." Proceedings of the MileTS 21 (2021).

    .. [ermshaus2023clasp] Ermshaus, Arik, Patrick Schäfer, and Ulf Leser. "ClaSP:
       parameter-free time series segmentation." Data Mining and Knowledge Discovery
       37.3 (2023): 1262-1300, doi: `10.1007/s10618-023-00923-x <https://doi.org/10.1007/s10618-023-00923-x>`_
    """
    # Check the input
    check_is_valid_window_size(window_size)
    if not utils.is_valid_array_like(X):
        raise ValueError("X must be a valid, numerical array-like")

    # Initialize the variable
    window_size_ = -1

    # If an int is given, then we can simply return the given window size
    if isinstance(window_size, int):
        return window_size

    # Check if the time series is univariate (error should not be raise if given window size is an integer)
    elif not utils.is_univariate(X):
        raise ValueError(
            "It only makes sens to compute the window size in univariate time series."
        )

    # Use the fft to compute a window size
    elif window_size == "fft":
        window_size_ = _dominant_fourier_frequency(
            X, lower_bound=lower_bound, upper_bound=upper_bound
        )

    # Use the acf to compute a window size
    elif window_size == "acf":
        window_size_ = _highest_autocorrelation(
            X, lower_bound=lower_bound, upper_bound=upper_bound
        )

    elif window_size == "mwf":
        window_size_ = _mwf(X, lower_bound=lower_bound, upper_bound=upper_bound)

    # Use SUSS to compute a window size
    elif window_size == "suss":
        window_size_ = _suss(X, lower_bound=lower_bound, threshold=threshold)

    # Check if a valid window size was computed, and raise an error if necessary
    if window_size_ == -1:
        if default_window_size is None:
            raise ValueError(
                f"Something went wrong when computing the window size using '{window_size}' on a time series with shape {X.shape}!"
            )
        else:
            return default_window_size
    else:
        return window_size_


def _dominant_fourier_frequency(
    X: np.ndarray, lower_bound: int, upper_bound: int
) -> int:
    # https://github.com/ermshaua/window-size-selection/blob/main/src/window_size/period.py#L10
    fourier = np.fft.fft(X)
    freq = np.fft.fftfreq(X.shape[0], 1)

    magnitudes = []
    window_sizes = []

    for coef, freq in zip(fourier, freq):
        if coef and freq > 0:
            window_size = int(1 / freq)
            mag = math.sqrt(coef.real * coef.real + coef.imag * coef.imag)

            if lower_bound <= window_size <= upper_bound:
                window_sizes.append(window_size)
                magnitudes.append(mag)

    if len(window_sizes) == 0:
        return -1

    return window_sizes[np.argmax(magnitudes)]


def _highest_autocorrelation(X: np.ndarray, lower_bound: int, upper_bound: int):
    # https://github.com/ermshaua/window-size-selection/blob/main/src/window_size/period.py#L29
    acf_values = acf(X, fft=True, nlags=int(X.shape[0] / 2))

    peaks, _ = scipy.signal.find_peaks(acf_values)
    peaks = peaks[np.logical_and(peaks >= lower_bound, peaks < upper_bound)]
    corrs = acf_values[peaks]

    if peaks.shape[0] == 0:
        return -1

    return peaks[np.argmax(corrs)]


def _mwf(X: np.ndarray, lower_bound: int, upper_bound: int) -> int:
    # https://github.com/ermshaua/window-size-selection/blob/main/src/window_size/mwf.py#L16

    def moving_mean(time_series: np.ndarray, w: int):
        moving_avg = np.cumsum(time_series, dtype=float)
        moving_avg[w:] = moving_avg[w:] - moving_avg[:-w]
        return moving_avg[w - 1 :] / w

    all_averages = []
    window_sizes = list(range(lower_bound, upper_bound))

    for window_size in window_sizes:
        all_averages.append(np.array(moving_mean(X, window_size)))

    moving_average_residuals = []
    for i in range(len(window_sizes)):
        moving_avg = all_averages[i][: len(all_averages[-1])]
        moving_avg_residual = np.log(abs(moving_avg - moving_avg.mean()).sum())
        moving_average_residuals.append(moving_avg_residual)

    b = (np.diff(np.sign(np.diff(moving_average_residuals))) > 0).nonzero()[
        0
    ] + 1  # local min

    if len(b) == 0:
        return -1
    if len(b) < 3:
        return window_sizes[b[0]]

    w = np.mean([window_sizes[b[i]] / (i + 1) for i in range(3)])
    return int(w)


def _suss(X: np.ndarray, lower_bound: int, threshold: float) -> int:
    # https://github.com/ermshaua/window-size-selection/blob/main/src/window_size/suss.py#L25
    # Implementation has been changed to remove pandas dependencies (in `suss_score`)

    def suss_score(time_series: np.ndarray, w: int):

        # Compute the statistics in each window
        windows = np.lib.stride_tricks.sliding_window_view(time_series, w)
        local_stats = np.array(
            [
                windows.mean(axis=1) - global_mean,
                windows.std(axis=1) - global_std,
                (windows.max(axis=1) - windows.min(axis=1)) - global_min_max,
            ]
        )

        # Compute Euclidean distance between local and global stats
        stats_diff = np.sqrt(np.sum(np.square(local_stats), axis=0)) / np.sqrt(w)
        return np.mean(stats_diff)

    if X.max() > X.min():
        X = (X - X.min()) / (X.max() - X.min())

    global_mean = np.mean(X)
    global_std = np.std(X)
    global_min_max = np.max(X) - np.min(X)

    max_suss_score = suss_score(X, 1)
    min_suss_score = suss_score(X, X.shape[0] - 1)
    if min_suss_score == max_suss_score:
        return -1

    # exponential search (to find window size interval)
    exp = 0
    while True:
        window_size = 2**exp

        if window_size < lower_bound:
            exp += 1
            continue

        score = 1 - (suss_score(X, window_size) - min_suss_score) / (
            max_suss_score - min_suss_score
        )

        if score > threshold:
            break

        exp += 1

    lbound, ubound = max(lower_bound, 2 ** (exp - 1)), min(2**exp + 1, X.shape[0] - 1)

    # binary search (to find window size in interval)
    while lbound <= ubound:
        window_size = int((lbound + ubound) / 2)
        score = 1 - (suss_score(X, window_size) - min_suss_score) / (
            max_suss_score - min_suss_score
        )

        if score < threshold:
            lbound = window_size + 1
        elif score > threshold:
            ubound = window_size - 1
        else:
            lbound = window_size
            break

    return 2 * lbound



import multiprocessing
import time
import tracemalloc
from functools import partial
from typing import Dict, List, Union

import numpy as np
import pandas as pd

from dtaianomaly.anomaly_detection.BaseDetector import BaseDetector, Supervision
from dtaianomaly.data.DataSet import DataSet
from dtaianomaly.data.LazyDataLoader import LazyDataLoader
from dtaianomaly.evaluation.metrics import BinaryMetric, Metric
from dtaianomaly.pipeline.EvaluationPipeline import EvaluationPipeline
from dtaianomaly.preprocessing.Preprocessor import Identity, Preprocessor
from dtaianomaly.thresholding.thresholding import Thresholding
from dtaianomaly.workflow.error_logging import log_error
from dtaianomaly.workflow.utils import (
    build_pipelines,
    convert_to_list,
    convert_to_proba_metrics,
)


class Workflow:
    """
    Run anomaly detection experiments

    Run all combinations of ``dataloaders``, ``preprocessors``, ``detectors``,
    and ``metrics``. The metrics requiring a thresholding operation are
    combined with every element of ``thresholds``. If an error occurs in any
    execution of an anomaly detector or loading of data, then the error will
    be written to an error file, which is an executable Python file to reproduce
    the error.

    Parameters
    ----------
    dataloaders: LazyDataLoader or list of LazyDataLoader
        The dataloaders that will be used to load data, and consequently
        this data is used for evaluation within this workflow.

    metrics: Metric or list of Metric
        The metrics to evaluate within this workflow.

    detectors: BaseDetector or list of BaseDetector
        The anomaly detectors to evaluate.

    thresholds: Thresholding or list of Thresholding, default=None
        The thresholds used for converting continuous anomaly scores to
        binary anomaly predictions. Each threshold will be combined with
        each :py:class:`~dtaianomaly.evaluation.BinaryMetric` given via
        the ``metrics`` parameter. The thresholds do not apply on a
        :py:class:`~dtaianomaly.evaluation.ProbaMetric`. If equals None
        or an empty list, then all the given metrics via the ``metrics``
        argument must be of type :py:class:`~dtaianomaly.evaluation.ProbaMetric`.
        Otherwise, a ValueError will be raised.

    preprocessors: Preprocessor or list of Preprocessor, default=None
        The preprocessors to apply before evaluating the model. If equals
        None or an empty list, then no preprocssing will be done, aka.
        using :py:class:`dtaianomaly.preprocessing.Preprocessor` as the
        preprocessor for each pipeline.

    n_jobs: int, default=1
        Number of processes to run in parallel while evaluating all
        combinations.

    trace_memory: bool, default=False
        Whether or not memory usage of each run is reported. While this
        might give additional insights into the models, their runtime
        will be higher due to additional internal bookkeeping.

    error_log_path: str, default='./error_logs'
        The path in which the error logs should be saved.

    fit_unsupervised_on_test_data: bool, default=False
        Whether to fit the unsupervised anomaly detectors on the test data.
        If True, then the test data will be used to fit the detector and
        to evaluate the detector. This is no issue, since unsupervised
        detectors do not use labels and can deal with anomalies in the
        training data.
    """

    dataloaders: List[LazyDataLoader]
    pipelines: List[EvaluationPipeline]
    provided_preprocessors: bool
    n_jobs: int
    trace_memory: bool
    error_log_path: str
    fit_unsupervised_on_test_data: bool

    def __init__(
        self,
        dataloaders: Union[LazyDataLoader, List[LazyDataLoader]],
        metrics: Union[Metric, List[Metric]],
        detectors: Union[BaseDetector, List[BaseDetector]],
        preprocessors: Union[Preprocessor, List[Preprocessor]] = None,
        thresholds: Union[Thresholding, List[Thresholding]] = None,
        n_jobs: int = 1,
        trace_memory: bool = False,
        error_log_path: str = "./error_logs",
        fit_unsupervised_on_test_data: bool = False,
    ):

        # Make sure the inputs are lists.
        dataloaders = convert_to_list(dataloaders)
        metrics = convert_to_list(metrics)
        thresholds = convert_to_list(thresholds or [])
        preprocessors = convert_to_list(preprocessors or [])
        self.provided_preprocessors = len(preprocessors) > 0
        if not self.provided_preprocessors:
            preprocessors = [Identity()]
        detectors = convert_to_list(detectors)

        # Add thresholding to the binary metrics
        if len(thresholds) == 0 and any(
            isinstance(metric, BinaryMetric) for metric in metrics
        ):
            raise ValueError(
                "There should be at least one thresholding option if a binary metric is passed!"
            )
        proba_metrics = convert_to_proba_metrics(metrics=metrics, thresholds=thresholds)

        # Perform checks on input
        if len(dataloaders) == 0:
            raise ValueError(
                "At least one data loader should be given to the workflow!"
            )
        if len(metrics) == 0:
            raise ValueError("At least one metrics should be given to the workflow!")
        if len(detectors) == 0:
            raise ValueError("At least one detectors should be given to the workflow!")
        if n_jobs < 1:
            raise ValueError("There should be at least one job within a workflow!")

        # Set the properties of this workflow
        self.pipelines = build_pipelines(
            preprocessors=preprocessors, detectors=detectors, metrics=proba_metrics
        )
        self.dataloaders = dataloaders
        self.n_jobs = n_jobs
        self.trace_memory = trace_memory
        self.error_log_path = error_log_path
        self.fit_unsupervised_on_test_data = fit_unsupervised_on_test_data

    def run(self) -> pd.DataFrame:
        """
        Run the experimental workflow. Evaluate each pipeline within this
        workflow on each dataset within this workflow in a grid-like manner.

        Returns
        -------
        results: pd.DataFrame
            A pandas dataframe with the results of this workflow. Each row
            represents an execution of an anomaly detector on a given dataset
            with some preprocessing steps. The columns correspond to the
            different evaluation metrics, running time and potentially also
            the memory usage.
        """
        # Create all the jobs
        unit_jobs = [
            (dataloader, pipeline)
            for dataloader in self.dataloaders
            for pipeline in self.pipelines
        ]

        # Execute the jobs
        if self.n_jobs == 1:
            result = [
                _single_job(
                    *job,
                    trace_memory=self.trace_memory,
                    error_log_path=self.error_log_path,
                    fit_unsupervised_on_test_data=self.fit_unsupervised_on_test_data,
                )
                for job in unit_jobs
            ]
        else:
            single_run_function = partial(
                _single_job,
                trace_memory=self.trace_memory,
                error_log_path=self.error_log_path,
                fit_unsupervised_on_test_data=self.fit_unsupervised_on_test_data,
            )
            with multiprocessing.Pool(processes=self.n_jobs) as pool:
                result = pool.starmap(single_run_function, unit_jobs)

        # Create a dataframe of the results
        results_df = pd.DataFrame(result)

        # Reorder the columns
        columns = [
            "Dataset",
            "Detector",
            "Preprocessor",
            "Runtime Fit [s]",
            "Runtime Predict [s]",
            "Runtime [s]",
        ]
        if self.trace_memory:
            columns.extend(
                ["Peak Memory Fit [MB]", "Peak Memory Predict [MB]", "Peak Memory [MB]"]
            )
        results_df = results_df[
            columns + [x for x in results_df.columns if x not in columns]
        ]

        # Drop the processors column, if none were provided.
        if not self.provided_preprocessors:
            results_df.drop(columns="Preprocessor", inplace=True)

        # Return the results
        return results_df


def _single_job(
    dataloader: LazyDataLoader,
    pipeline: EvaluationPipeline,
    trace_memory: bool,
    error_log_path: str,
    fit_unsupervised_on_test_data: bool,
) -> Dict[str, Union[str, float]]:
    # Initialize the results, and by default everything went wrong ('Error')
    results = {"Dataset": str(dataloader)}
    for key in pipeline.metrics + [
        "Detector",
        "Preprocessor",
        "Runtime Fit [s]",
        "Runtime Predict [s]",
        "Runtime [s]",
    ]:
        results[str(key)] = "Error"
    if trace_memory:
        for key in [
            "Peak Memory Fit [MB]",
            "Peak Memory Predict [MB]",
            "Peak Memory [MB]",
        ]:
            results[key] = "Error"

    # Try to load the data set, if this fails, return the results
    try:
        data_set = dataloader.load()
    except Exception as exception:
        results["Error file"] = log_error(error_log_path, exception, dataloader)
        return results

    # We can already save the used preprocessor and detector
    results["Preprocessor"] = str(pipeline.pipeline.preprocessor)
    results["Detector"] = str(pipeline.pipeline.detector)

    # Check if the dataset and the anomaly detector are compatible
    if not data_set.is_compatible(pipeline.pipeline):
        error_message = (
            f"Not compatible: detector with supervision {pipeline.pipeline.supervision} "
            f"for data set with compatible supervision ["
        )
        error_message += ", ".join([str(s) for s in data_set.compatible_supervision()])
        error_message += "]"
        for key, value in results.items():
            if value == "Error":
                results[key] = error_message
        return results

    # Format X_train, y_train, X_test and y_test
    X_test, y_test, X_train, y_train, fit_on_X_train = _get_train_test_data(
        data_set, pipeline.pipeline, fit_unsupervised_on_test_data
    )

    # Run the anomaly detector, and catch any exceptions
    try:
        # Fitting
        _start_tracing_memory(trace_memory)
        start = _start_tracing_runtime()
        pipeline.fit(X_train, y_train)
        results["Runtime Fit [s]"] = _end_tracing_runtime(start)
        _end_tracing_memory(trace_memory, results, "Peak Memory Fit [MB]")

        # Predicting
        _start_tracing_memory(trace_memory)
        start = _start_tracing_runtime()
        y_pred = pipeline.predict(X_test)
        results["Runtime Predict [s]"] = _end_tracing_runtime(start)
        _end_tracing_memory(trace_memory, results, "Peak Memory Predict [MB]")

        # Scoring
        y_test_ = pipeline.format_y_test(X_test, y_test)
        results.update(pipeline.evaluate(y_test_, y_pred))

        # Aggregate the used resources
        results["Runtime [s]"] = (
            results["Runtime Fit [s]"] + results["Runtime Predict [s]"]
        )
        if trace_memory:
            results["Peak Memory [MB]"] = max(
                results["Peak Memory Fit [MB]"], results["Peak Memory Predict [MB]"]
            )

    except Exception as exception:
        # Log the errors
        results["Error file"] = log_error(
            error_log_path, exception, dataloader, pipeline.pipeline, fit_on_X_train
        )

    # Return the results
    return results


def _start_tracing_runtime() -> float:
    return time.time()


def _end_tracing_runtime(start_time: float) -> float:
    return time.time() - start_time


def _start_tracing_memory(trace_memory: bool) -> None:
    if trace_memory:
        tracemalloc.start()


def _end_tracing_memory(trace_memory: bool, results, key) -> None:
    if trace_memory:
        _, peak = tracemalloc.get_traced_memory()
        results[key] = peak / 10**6
        tracemalloc.stop()


def _get_train_test_data(
    data_set: DataSet, detector: BaseDetector, fit_unsupervised_on_test_data: bool
) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, bool):
    """
    Separates the train and test data depending on the type of the anomaly
    detector and whether the test data should be used for fitting in an
    unsupervised detector.

    Also returns a bool indicating if the train data is actually used for
    fitting or not.
    """
    X_test = data_set.X_test
    y_test = data_set.y_test
    X_train = data_set.X_train
    y_train = data_set.y_train

    fit_on_X_train = True

    # If no train data is given but the detector is unsupervised, then use the test data for training
    # This is only ok if the detector is unsupervised, because no labels are used
    # If this happens, the train labels will be None anyway (otherwise data_set would be invalid)
    if detector.supervision == Supervision.UNSUPERVISED and X_train is None:
        X_train = X_test
        fit_on_X_train = False

    # If unsupervised detectors should fit on the test data.
    if (
        fit_unsupervised_on_test_data
        and detector.supervision == Supervision.UNSUPERVISED
    ):
        X_train = X_test
        fit_on_X_train = False

    return X_test, y_test, X_train, y_train, fit_on_X_train



"""
This module contains functions for plotting time series. It can be imported as follows:

>>> from dtaianomaly import visualization

The functions within this module offer alternative manners to nicely plot the time series
along with the ground truth or predicted anomalies.
"""

from .visualization import (
    format_time_steps,
    plot_anomaly_scores,
    plot_demarcated_anomalies,
    plot_time_series_anomalies,
    plot_time_series_colored_by_score,
    plot_with_zoom,
)

__all__ = [
    "plot_time_series_colored_by_score",
    "plot_time_series_anomalies",
    "plot_demarcated_anomalies",
    "plot_with_zoom",
    "plot_anomaly_scores",
    "format_time_steps",
]



import numpy as np

from dtaianomaly.evaluation.metrics import BinaryMetric, ProbaMetric


class BestThresholdMetric(ProbaMetric):
    """
    Compute the maximum score of a binary metric over all thresholds.
    This method will iterate over the possible threshold for given
    predicted anomaly scores, compute the binary metric for each
    threshold, and then return the score for the highest threshold.

    Parameters
    ----------
    metric: BinaryMetric
        Instance of the desired `Metric` class
    max_nb_thresholds: int, default=-1
        The maximum number of thresholds to use for computing the best threshold.
        If ``max_nb_thresholds = -1``, all thresholds will be used. Otherwise, the
        value indicates the subsample of all possible thresholds that should be used.
        This subset is created by first sorting the possible unique thresholds, and
        then selecting the threshold at regular intervals (i.e., the 3rd, 6th, 9th, ...).
        We recommend using the default value (use all thresholds), but can be used
        for reducing the resource requirements.

    Attributes
    ----------
    threshold_: float
        The threshold resulting in the best performance.
    """

    metric: BinaryMetric
    max_nb_thresholds: int
    threshold_: float

    def __init__(self, metric: BinaryMetric, max_nb_thresholds: int = -1) -> None:
        if not isinstance(metric, BinaryMetric):
            raise TypeError(f"metric expects 'BinaryMetric', got {type(metric)}")
        if not isinstance(max_nb_thresholds, int) or isinstance(
            max_nb_thresholds, bool
        ):
            raise TypeError("`max_nb_thresholds` should be an integer")
        if max_nb_thresholds <= 0 and max_nb_thresholds != -1:
            raise ValueError(
                "`max_nb_thresholds` must be strictly positive or equal to -1!"
            )
        self.metric = metric
        self.max_nb_thresholds = max_nb_thresholds

    def _compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:

        # Sort all the predicted scores
        sorted_predicted_scores = np.sort(np.unique(y_pred))

        # Get all possible thresholds
        thresholds = (sorted_predicted_scores[:-1] + sorted_predicted_scores[1:]) / 2.0

        # Add the minimum and maximum threshold
        thresholds = np.append(np.insert(thresholds, 0, 0), 1)

        # Select a subset of the thresholds, if requested and useful
        if 0 < self.max_nb_thresholds < thresholds.shape[0]:
            selected_thresholds = np.linspace(
                0, thresholds.shape[0], self.max_nb_thresholds + 2, dtype=int
            )[1:-1]
            thresholds = thresholds[selected_thresholds]

        # Compute the score for each threshold
        scores = [
            self.metric._compute(y_true, y_pred >= threshold)
            for threshold in thresholds
        ]

        # Get the best score and the corresponding threshold
        i = np.argmax(scores)
        best_score = scores[i]
        self.threshold_ = thresholds[i]

        # Return the best score
        return best_score



from typing import Optional, Tuple

import numpy as np
from sklearn.exceptions import NotFittedError

from dtaianomaly.preprocessing.Preprocessor import Preprocessor


class MinMaxScaler(Preprocessor):
    """
    Rescale raw time series to a [0, 1] via min-max scaling. The
    minimum and maximum is computed on a training set, after which
    these values can be used to transform a new time series. Therefore,
    there is no guarantee that the values of the transformed test set
    will actually be in the range [0, 1].

    For multivariate time series, each attribute will be normalized
    independently, i.e., the minimum and maximum of each attribute in
    the transformed time series will 0 and 1, respectively.

    If the minimum and maximum of an attribute is the same (the time
    series consists of only one value), then the transformation will
    not do anything.

    Attributes
    ----------
    min_: array-like of shape (n_attributes)
        The minimum value in each attribute of the training data.
    max_: array-like of shape (n_attributes)
        The maximum value in each attribute of the training data.

    Raises
    ------
    NotFittedError
        If the `transform` method is called before fitting this MinMaxScaler.
    """

    min_: np.array
    max_: np.array

    def _fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "MinMaxScaler":
        if len(X.shape) == 1 or X.shape[1] == 1:
            # univariate case
            self.min_ = np.array([np.nanmin(X)])
            self.max_ = np.array([np.nanmax(X)])
        else:
            # multivariate case
            self.min_ = np.nanmin(X, axis=0)
            self.max_ = np.nanmax(X, axis=0)

        # Adjust to deal with constant attributes
        constant_attributes = self.min_ == self.max_
        self.min_ = np.where(constant_attributes, 0, self.min_)
        self.max_ = np.where(constant_attributes, 1, self.max_)

        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        if not (hasattr(self, "min_") and hasattr(self, "max_")):
            raise NotFittedError(f"Call `fit` before using transform on {str(self)}")
        if not (
            (len(X.shape) == 1 and self.min_.shape[0] == 1)
            or X.shape[1] == self.min_.shape[0]
        ):
            raise AttributeError(
                f"Trying to min max scale a time series with {X.shape[0]} attributes while it was fitted on {self.min_.shape[0]} attributes!"
            )

        X_ = (X - self.min_) / (self.max_ - self.min_)
        return X_, y




import pytest
import numpy as np
from typing import Any

from dtaianomaly.utils import is_valid_list, is_valid_array_like, is_univariate, get_dimension


class TestIsValidList:

    def test_empty_list(self):
        assert is_valid_list([], Any)

    def test_np_array(self):
        assert not is_valid_list(np.array([1, 2, 3, 4, 5, 6]), Any)

    def test_valid(self):
        assert is_valid_list([1, 2, 3, 4, 5, 6], int)

    def test_invalid(self):
        assert not is_valid_list([1, 2, 3, '4', 5, 6], int)


class TestIsValidArrayLike:

    def test_empty_list(self):
        assert is_valid_array_like([])

    def test_empty_np_array(self):
        assert is_valid_array_like(np.array([]))

    def test_valid_list_int(self):
        assert is_valid_array_like([1, 2, 3, 4, 5])

    def test_valid_list_int_2d(self):
        assert is_valid_array_like([[1], [2], [3], [4], [5]])

    def test_valid_np_array_int(self):
        assert is_valid_array_like(np.array([1, 2, 3, 4, 5]))

    def test_valid_np_array_int_2d(self):
        assert is_valid_array_like(np.array([[1], [2], [3], [4], [5]]))

    def test_valid_list_float(self):
        assert is_valid_array_like([1.9, 2.8, 3.7, 4.6, 5.5])

    def test_valid_np_array_float(self):
        assert is_valid_array_like(np.array([1.9, 2.8, 3.7, 4.6, 5.5]))

    def test_valid_list_bool(self):
        assert is_valid_array_like([True, True, False, True, False])

    def test_valid_np_array_bool(self):
        assert is_valid_array_like(np.array([True, True, False, True, False]))

    def test_valid_list_mixed_type(self):
        assert is_valid_array_like([1.9, 2, True, 4, 5.5])

    def test_valid_np_array_mixed_type(self):
        assert is_valid_array_like(np.array([1.9, 2, True, 4, 5.5]))

    def test_valid_multivariate_list_int(self):
        assert is_valid_array_like([[1, 10], [2, 20], [3, 30], [4, 40], [5, 50]])

    def test_valid_multivariate_np_array_int(self):
        assert is_valid_array_like(np.array([[1, 10], [2, 20], [3, 30], [4, 40], [5, 50]]))

    def test_valid_multivariate_list_float(self):
        assert is_valid_array_like([[1.9, 9.1], [2.8, 8.2], [3.7, 7.3], [4.6, 6.4], [5.5, 5.5]])

    def test_valid_multivariate_np_array_float(self):
        assert is_valid_array_like(np.array([[1.9, 9.1], [2.8, 8.2], [3.7, 7.3], [4.6, 6.4], [5.5, 5.5]]))

    def test_valid_multivariate_list_bool(self):
        assert is_valid_array_like([[True, True], [True, False], [False, False], [True, False], [False, True]])

    def test_valid_multivariate_np_array_bool(self):
        assert is_valid_array_like(np.array([[True, True], [True, False], [False, False], [True, False], [False, True]]))

    def test_valid_multivariate_list_mixed_type(self):
        assert is_valid_array_like([[1.9, 1], [2, False], [True, 3], [4, 6.4], [5.5, True]])

    def test_valid_multivariate_np_array_mixed_type(self):
        assert is_valid_array_like(np.array([[1.9, 1], [2, False], [True, 3], [4, 6.4], [5.5, True]]))

    def test_invalid_list(self):
        assert not is_valid_array_like([1, 2, 3, 4, '5'])

    def test_invalid_np_array(self):
        assert not is_valid_array_like(np.array([1, 2, 3, 4, '5']))

    def test_invalid_int(self):
        assert not is_valid_array_like(1)

    def test_invalid_float(self):
        assert not is_valid_array_like(1.9)

    def test_invalid_str(self):
        assert not is_valid_array_like('1')

    def test_invalid_bool(self):
        assert not is_valid_array_like(True)

    def test_invalid_none(self):
        assert not is_valid_array_like(None)

    def test_invalid_multivariate_list_type(self):
        assert not is_valid_array_like([[1, 10], [2, 20], [3, 30], [4, 40], [5, '50']])

    def test_invalid_multivariate_with_str(self):
        assert not is_valid_array_like([[1, 10], [2, 20], [3, 30], [4, 40], '55'])

    def test_invalid_multivariate_list_dimension(self):
        assert not is_valid_array_like([[1, 10], [2, 20], [3, 30], [4, 40, 400], [5, 50]])

    def test_invalid_multivariate_list_non_list(self):
        assert not is_valid_array_like([[1], [2], 3, [4], [5]])

    def test_invalid_multivariate_list_first_non_list(self):
        assert not is_valid_array_like([1, [2], [3], [4], [5]])


class TestIsUnivariate:

    def test_multivariate(self, multivariate_time_series):
        assert not is_univariate(multivariate_time_series)

    def test_univariate_1_dimension(self, univariate_time_series):
        univariate_time_series = univariate_time_series.squeeze()
        assert len(univariate_time_series.shape) == 1
        assert is_univariate(univariate_time_series)

    def test_univariate_2_dimensions(self, univariate_time_series):
        univariate_time_series = univariate_time_series.reshape(univariate_time_series.shape[0], 1)
        assert len(univariate_time_series.shape) == 2
        assert is_univariate(univariate_time_series)

    def test_multivariate_list(self):
        assert not is_univariate([[0, 0], [1, 10], [2, 20], [3, 30], [4, 40], [5, 50]])

    def test_univariate_list(self, univariate_time_series):
        assert is_univariate([i for i in univariate_time_series])


class TestGetDimension:

    def test_single_dimension(self):
        rng = np.random.default_rng(42)
        X = np.random.uniform(size=1000)
        assert get_dimension(X) == 1

    @pytest.mark.parametrize('dimension', [1, 2, 3, 5, 10])
    def test_dimension(self, dimension):
        rng = np.random.default_rng(42)
        X = np.random.uniform(size=(1000, dimension))
        assert get_dimension(X) == dimension



"""
This module contains functionality to dynamically load data when
executing a pipeline or workflow. It can be imported as follows:

>>> from dtaianomaly import data

Custom data loaders can be implemented by extending :py:class:`~dtaianomaly.data.LazyDataLoader`.
"""

from .DataSet import DataSet
from .LazyDataLoader import LazyDataLoader, from_directory
from .synthetic import demonstration_time_series, make_sine_wave
from .UCRLoader import UCRLoader

__all__ = [
    "LazyDataLoader",
    "DataSet",
    "from_directory",
    "demonstration_time_series",
    "make_sine_wave",
    "UCRLoader",
]



from pyod.models.kpca import KPCA

from dtaianomaly.anomaly_detection.BaseDetector import Supervision
from dtaianomaly.anomaly_detection.PyODAnomalyDetector import PyODAnomalyDetector


class KernelPrincipalComponentAnalysis(PyODAnomalyDetector):
    """
    Anomaly detector based on the Kernel Principal Component Analysis (KPCA).

    Standard PCA maps the data to a lower dimensional space through linear
    projections. Deviations in this lower dimensional space are then
    considered to be anomalies. KPCA [hoffmann2007kernel]_ is a non-linear
    extension of PCA, which maps the data into a new kernel space, from
    which the principal components are learned.

    Notes
    -----
    KPCA inherets from :py:class:`~dtaianomaly.anomaly_detection.PyodAnomalyDetector`.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs:
        Arguments to be passed to the PyOD PCA.

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : KPCA
        A KPCA-detector of PyOD

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import KernelPrincipalComponentAnalysis
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> kpca = KernelPrincipalComponentAnalysis(10, n_components=2).fit(x)
    >>> kpca.decision_function(x)
    array([0.03151377, 0.03697829, 0.04415575, ..., 0.03345565, 0.0330048 ,
           0.03089501])

    References
    ----------
    .. [hoffmann2007kernel] Heiko Hoffmann. Kernel pca for novelty detection. Pattern recognition,
       40(3):863–874, 2007, doi: `10.1016/j.patcog.2006.07.009 <https://doi.org/10.1016/j.patcog.2006.07.009>`_.
    """

    def _initialize_detector(self, **kwargs) -> KPCA:
        return KPCA(**kwargs)

    def _supervision(self):
        return Supervision.SEMI_SUPERVISED



"""
This module contains the workflow functionality.

>>> from dtaianomaly import workflow

Below we illustrate how a simple workflow can be initialized, which will
apply Matrix Profile and Isolation Forest on a dataset from the UCR
archive, and compute the area under the ROC and PR curves:

.. testsetup::

   import os
   os.chdir('..')


>>> from dtaianomaly.data import UCRLoader
>>> from dtaianomaly.anomaly_detection import MatrixProfileDetector, IsolationForest
>>> from dtaianomaly.evaluation import AreaUnderROC, AreaUnderPR
>>> workflow = workflow.Workflow(
...     dataloaders=[
...         UCRLoader(path='data/UCR-time-series-anomaly-archive/001_UCR_Anomaly_DISTORTED1sddb40_35000_52000_52620.txt'),
...     ],
...     detectors=[MatrixProfileDetector(window_size=100), IsolationForest(15)],
...     metrics=[AreaUnderROC(), AreaUnderPR()]
... )

We refer to the `documentation <https://dtaianomaly.readthedocs.io/en/stable/getting_started/quantitative_evaluation.html>`_
for more information regarding the configuration and use of a Workflow.
"""

from .Workflow import Workflow
from .workflow_from_config import interpret_config, workflow_from_config

__all__ = ["Workflow", "workflow_from_config", "interpret_config"]



from typing import Optional, Tuple

import numpy as np

from dtaianomaly.preprocessing.Preprocessor import Preprocessor


class PiecewiseAggregateApproximation(Preprocessor):
    """
    Performs piecewise aggregate approximation.

    Piecewise Aggregate Approximation (PAA) [keogh2001dimensionality]_ is a
    form of dimensionality reduction of time series, originally proposed for
    fast indexing of time series in large databases. Given a value for :math:`n`,
    PAA divides the time series in :math:`n` equi-sized frames. Next, each frame
    is replaced by its mean value. Specifically, for a time series :math:`x` of
    length :math:`N`, position :math:`i` in the transformed time series :math:`y`
    equals:

    .. math::

       y_i = \\frac{n}{N} \\displaystyle\\sum_{j=N/N(i-1)+1}^{(n/N)i} x_j

    For multivariate time series, the dimension of each attribute is reduced
    independently, but the same frames are used.

    Parameters
    ----------
    n: int
        The number of equi-sized frames to generate.

    References
    ----------
    .. [keogh2001dimensionality] Keogh, E., Chakrabarti, K., Pazzani, M. et al.
       Dimensionality Reduction for Fast Similarity Search in Large Time Series
       Databases. Knowledge and Information Systems 3, 263–286 (2001).
       doi: `10.1007/PL00011669 <https://doi.org/10.1007/PL00011669>`_.
    """

    n: int

    def __init__(self, n: int):
        super().__init__()

        if not isinstance(n, int) or isinstance(n, bool):
            raise TypeError("`n` should be an integer")
        if n <= 0:
            raise ValueError("'n' must be strictly positive!")

        self.n = n

    def _fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "Preprocessor":
        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        if X.shape[0] <= self.n:
            return X, y

        X_ = paa(X, self.n)
        if y is None:
            return X_, y
        else:
            return X_, np.where(paa(y, self.n) < 0.5, 0, 1)


def paa(x: np.ndarray, n: int) -> np.ndarray:
    indices = np.linspace(0, x.shape[0], n + 1, dtype=int, endpoint=True)
    return np.array([np.mean(x[s:e], axis=0) for s, e in zip(indices, indices[1:])])



"""
This module contains functionality to evaluate performance of an
anomaly detector. It can be imported as follows:

>>> from dtaianomaly import evaluation

Custom evaluation metrics can be implemented by extending :py:class:`~dtaianomaly.evaluation.Metric` or
:py:class:`~dtaianomaly.evaluation.ProbaMetric`. The former expects predicted "decisions" (anomaly or not),
the latter predicted "scores" (more or less anomalous). This distinction is important for later use in
a :py:class:`~dtaianomaly.workflow.Worfklow`.
"""

from .BestThresholdMetric import BestThresholdMetric
from .metrics import BinaryMetric, Metric, ProbaMetric, ThresholdMetric
from .point_adjusted_binary_metrics import (
    PointAdjustedFBeta,
    PointAdjustedPrecision,
    PointAdjustedRecall,
)
from .simple_binary_metrics import FBeta, Precision, Recall
from .simple_proba_metrics import AreaUnderPR, AreaUnderROC

__all__ = [
    "Metric",
    "BinaryMetric",
    "ProbaMetric",
    "ThresholdMetric",
    "Precision",
    "Recall",
    "FBeta",
    "AreaUnderPR",
    "AreaUnderROC",
    "PointAdjustedPrecision",
    "PointAdjustedRecall",
    "PointAdjustedFBeta",
    "BestThresholdMetric",
]



import numpy as np
from sklearn import metrics

from dtaianomaly.evaluation.metrics import BinaryMetric


class Precision(BinaryMetric):
    """
    Computes the Precision score.

    Precision measures how accurately the model identifies anomalies.
    It reflects the proportion of detected anomalies that are truly
    abnormal. This is particularly important when the cost of false
    positives (normal events incorrectly flagged as anomalies) is high.

    Mathematically, precision is the ratio of true positives (correctly
    identified anomalies) to all predicted positives, which includes
    both true anomalies and false positives (normal events mistakenly
    flagged as anomalies). It can be expressed as:

    .. math::

       \\text{Precision} = \\frac{\\text{True Anomalies}}{\\text{True Anomalies} + \\text{False Positives}}


    A high precision in anomaly detection indicates that the model
    generates few false alarms, ensuring that most flagged anomalies
    are truly abnormal events. However, it does not measure how many
    anomalies were actually identified.
    """

    def __init__(self) -> None:
        super().__init__()

    def _compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        return metrics.precision_score(y_true=y_true, y_pred=y_pred)


class Recall(BinaryMetric):
    """
    Computes the Recall score.

    Recall measures the model's ability to correctly identify all actual
    anomalies. It tells us the proportion of true anomalies that were
    successfully detected by the model. In an anomaly detection system,
    recall answers the question: "Of all the anomalies that occurred,
    how many did the model detect?" A high recall is especially important
    when missing actual anomalies (false negatives) could have severe
    consequences.

    Mathematically, recall is the ratio of true positives (correctly
    identified anomalies) to all actual positives, which includes both
    true anomalies and false negatives (missed anomalies). It can be expressed as:

    .. math::

       \\text{Recall} = \\frac{\\text{True Anomalies}}{\\text{True Anomalies} + \\text{False Negatives}}

    A high recall ensures that most anomalies are detected, but it doesn’t
    account for how many false positives (normal events incorrectly flagged
    as anomalies) were generated, which is handled by precision.
    """

    def __init__(self) -> None:
        super().__init__()

    def _compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        return metrics.recall_score(y_true=y_true, y_pred=y_pred)


class FBeta(BinaryMetric):
    """
    Computes the :math:`F_\\beta` score.

    The :math:`F_\\beta` combines both precision and recall into a single
    value. It provides a balanced evaluation of a model’s performance,
    especially in anomaly detection, where there is often a trade-off
    between catching all anomalies (high recall) and minimizing false
    alarms (high precision). The parameter :math:`\\beta` controls the balance
    between precision and recall. A :math:`\\beta > 1` gives more weight to
    recall, useful when missing anomalies is costly, while :math:`\\beta < 1`
    emphasizes precision, reducing false positives.

    The :math:`F_\\beta` score is the harmonic mean of precision and recall.
    It can be expressed as:

    .. math::

       F_\\beta = \\frac{(1 + \\beta^2) \\text{tp}}
                        {(1 + \\beta^2) \\text{tp} + \\text{fp} + \\beta^2 \\text{fn}}

    A high :math:`F_\\beta` score indicates a good balance between detecting
    actual anomalies and minimizing false positives.

    Parameters
    ----------
    beta: int, float, default=1
        Desired beta parameter.

    See also
    --------
    Precision: Compute the Precision score.
    Recall: Compute the Recall score.
    """

    beta: float

    def __init__(self, beta: (float, int) = 1) -> None:
        if not isinstance(beta, (int, float)) or isinstance(beta, bool):
            raise TypeError("`beta` should be numeric")
        if beta <= 0.0:
            raise ValueError("`beta` should be strictly positive")

        super().__init__()
        self.beta = beta

    def _compute(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        return metrics.fbeta_score(y_true=y_true, y_pred=y_pred, beta=self.beta)



"""
This module contains functionality to combine preprocessing, anomaly detection,
and evaluation in a single wrapped object.

>>> from dtaianomaly import pipeline

Users are not expected to extend the base Pipeline objects as they are
wrappers of underlying ``dtaianomaly`` objects. Custom functionality is
better achieved by implementing the :py:class:`dtaianomaly.preprocessing.Preprocessor`,
:py:class:`dtaianomaly.anomaly_detection.BaseDetector` or
:py:class:`dtaianomaly.evaluation.Metric` objects.
"""

from .EvaluationPipeline import EvaluationPipeline
from .Pipeline import Pipeline

__all__ = ["Pipeline", "EvaluationPipeline"]



from typing import Optional, Tuple

import numpy as np
from sklearn.exceptions import NotFittedError

from dtaianomaly.preprocessing.Preprocessor import Preprocessor


class StandardScaler(Preprocessor):
    """
    Standard scale the data: rescale to zero mean, unit variance.

    Rescale to zero mean and unit variance. A mean value and standard
    deviation is computed on a training set, after which these values
    can be used to transform a new time series. Therefore, there is no
    guarantee that the values of the transformed test set will actually
    have zero mean and unit variance.

    For multivariate time series, each attribute will be normalized
    independently, i.e., the mean and std of each attribute in the
    transformed time series will 1.0 and 0.0, respectively.

    Parameters
    ----------
    min_std: float, default = 1e-9
        The minimum std required to actually Z-normalize an attribute.
        If the standard deviation is below this value, then no normalization
        will be applied. This prevents amplifying noise in the data.

    Attributes
    ----------
    mean_: array-like of shape (n_attributes)
        The mean value in each attribute of the training data.
    std_: array-like of shape (n_attributes)
        The standard deviation in each attribute of the training data.

    Raises
    ------
    NotFittedError
        If the `transform` method is called before fitting this StandardScaler.
    """

    min_std: float
    mean_: np.array
    std_: np.array

    def __init__(self, min_std: float = 1e-9):
        self.min_std = min_std

    def _fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "StandardScaler":
        if len(X.shape) == 1 or X.shape[1] == 1:
            # univariate case
            self.mean_ = np.array([np.nanmean(X)])
            self.std_ = np.array([np.nanstd(X)])
        else:
            # multivariate case
            self.mean_ = np.nanmean(X, axis=0)
            self.std_ = np.nanstd(X, axis=0)

        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        if not (hasattr(self, "mean_") and hasattr(self, "std_")):
            raise NotFittedError(f"Call `fit` before using transform on {str(self)}")
        if not (
            (len(X.shape) == 1 and self.mean_.shape[0] == 1)
            or X.shape[1] == self.mean_.shape[0]
        ):
            raise AttributeError(
                f"Trying to standard scale a time series with {X.shape[0]} attributes while it was fitted on {self.mean_.shape[0]} attributes!"
            )

        # If the std of all attributes is 0, then no transformation happens
        if np.all((self.std_ < self.min_std)):
            X_ = X

        # Else, each attribute is normalized independently, except for the
        # attributes with 0 std.
        else:
            X_ = (X - self.mean_) / self.std_
            for i, (std) in enumerate(self.std_):
                if std < self.min_std:
                    X_[:, i] = X[:, i]

        return X_, y



from typing import Optional, Tuple

import numpy as np

from dtaianomaly.preprocessing.Preprocessor import Preprocessor


class MovingAverage(Preprocessor):
    """
    Computes the moving average of a time series. This is the unweighted
    average of the observations within a window.

    To compute the moving average at time :math:`t`, the window is centered at
    position :math:`t`. For an odd window size, the number of measurements taken
    before and after :math:`t` is equal (namely ``(window_size - 1 ) / 2``. For an
    even window size, there is one additional observation taken before :math:`t`,
    to ensure a correct window size.

    For multivariate time series, the moving average is computed within
    each attribute independently.

    Parameters
    ----------
    window_size: int
        Length of the window in which the average should be computed.
    """

    window_size: int

    def __init__(self, window_size: int) -> None:
        if window_size <= 0:
            raise ValueError("Window size must be strictly positive")
        self.window_size = window_size

    def _fit(self, X: np.ndarray, y: Optional[np.ndarray] = None) -> "MovingAverage":
        return self

    def _transform(
        self, X: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        # Add nan values at the beginning and end of the given array
        extend_front = np.full(self.window_size // 2, np.nan)
        extend_back = np.full(
            self.window_size // 2 - (self.window_size % 2 == 0), np.nan
        )
        if len(X.shape) == 2:
            extend_front = np.repeat(extend_front, X.shape[1]).reshape(-1, X.shape[1])
            extend_back = np.repeat(extend_back, X.shape[1]).reshape(-1, X.shape[1])
        X_extended = np.concatenate([extend_front, X, extend_back], axis=0)
        # Compute the average within each window
        X_ = np.array(
            [
                np.nanmean(window, axis=-1)
                for window in np.lib.stride_tricks.sliding_window_view(
                    X_extended, self.window_size, axis=0
                )
            ]
        )
        # Return the results
        return X_, y



import abc
from typing import Optional, Union

import numpy as np
from pyod.models.base import BaseDetector as PyODBaseDetector
from sklearn.exceptions import NotFittedError

from dtaianomaly import utils
from dtaianomaly.anomaly_detection.BaseDetector import BaseDetector, Supervision
from dtaianomaly.anomaly_detection.windowing_utils import (
    check_is_valid_window_size,
    compute_window_size,
    reverse_sliding_window,
    sliding_window,
)


class PyODAnomalyDetector(BaseDetector, abc.ABC):
    """
    Abstract class for anomaly detection based on the PyOD library.

    PyOD [zhao2019pyod]_ is a Python library for detecting anomalies in multivariate
    data. The anomaly detectors in PyOD typically deal with tabular data, which assumes
    i.i.d (independent and identically distributed) data. This is generally not the
    case for time series data, which has a temporal dependency. Nevertheless, the detectors
    of PyOD can be used for detecting anomalies in time series data.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs:
        Arguments to be passed to pyod anomaly detector

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : SklearnLocalOutlierFactor
        The PyOD anomaly detector

    References
    ----------
    .. [zhao2019pyod] Zhao, Y., Nasrullah, Z. and Li, Z., 2019. PyOD: A Python Toolbox
       for Scalable Outlier Detection. Journal of machine learning research (JMLR), 20(96),
       pp.1-7.
    """

    window_size: Union[int, str]
    stride: int
    kwargs: dict
    window_size_: int
    pyod_detector_: PyODBaseDetector

    def __init__(self, window_size: Union[str, int], stride: int = 1, **kwargs):
        super().__init__(self._supervision())

        check_is_valid_window_size(window_size)

        if not isinstance(stride, int) or isinstance(stride, bool):
            raise TypeError("`stride` should be an integer")
        if stride < 1:
            raise ValueError("`stride` should be strictly positive")

        self.window_size = window_size
        self.stride = stride
        self.kwargs = kwargs

        # Check if the PyOD detector can be correctly initialized
        self._initialize_detector(**self.kwargs)

    @abc.abstractmethod
    def _initialize_detector(self, **kwargs) -> PyODBaseDetector:
        """
        Initialize the PyOD anomaly detector.

        Parameters
        ----------
        kwargs:
            The hyperparameters to be passed to the PyOD anomaly detector.

        Returns
        -------
        A PyOD anomaly detector with the given hyperparameters.
        """

    @abc.abstractmethod
    def _supervision(self) -> Supervision:
        """
        Return the supervision of this anomaly detector.

        Returns
        -------
        supervision: Supervision
            The supervision of this PyOD anomaly detector.
        """

    def fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None, **kwargs
    ) -> "BaseDetector":
        """
        Fit this PyOD anomaly detector on the given data.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.
        y: ignored
            Not used, present for API consistency by convention.
        kwargs:
            Additional parameters to be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.

        Returns
        -------
        self: PyODAnomalyDetector
            Returns the instance itself

        Raises
        ------
        ValueError
            If `X` is not a valid array.
        """
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")

        X = np.asarray(X)
        self.window_size_ = compute_window_size(X, self.window_size, **kwargs)
        self.pyod_detector_ = self._initialize_detector(**self.kwargs)
        self.pyod_detector_.fit(sliding_window(X, self.window_size_, self.stride))

        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        """
        Compute decision scores.

        Parameters
        ----------
        X: array-like of shape (n_samples, n_attributes)
            Input time series.

        Returns
        -------
        decision_scores: array-like of shape (n_samples)
            The decision scores of the anomaly detector. Higher indicates more anomalous.

        Raises
        ------
        ValueError
            If `X` is not a valid array.
        NotFittedError
            If this method is called before fitting the anomaly detector.
        """
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")
        if not hasattr(self, "pyod_detector_") or not hasattr(self, "window_size_"):
            raise NotFittedError("Call the fit function before making predictions!")

        X = np.asarray(X)
        per_window_decision_scores = self.pyod_detector_.decision_function(
            sliding_window(X, self.window_size_, self.stride)
        )
        decision_scores = reverse_sliding_window(
            per_window_decision_scores, self.window_size_, self.stride, X.shape[0]
        )

        return decision_scores



from collections.abc import Sequence

import numpy as np


def is_valid_list(value, target_type) -> bool:
    """
    Check if the given list is a valid, with each instance being a member
    of the given type.

    Parameters
    ----------
    value: object
        The value to check if it is a valid list
    target_type: Type
        The type of each object in the given list

    Returns
    -------
    is_valid: bool
        True if and only if the given ``value`` is a list and all elements in
        the list are of type ``Type``, otherwise False.
    """
    return isinstance(value, list) and all(
        isinstance(item, target_type) for item in value
    )


def is_valid_array_like(array) -> bool:
    """
    Check if input is "array-like". Within ``dtaianomaly``, this is
    either a numpy array of numerical values or a python sequence of
    numerical values.

    Parameters
    ----------
    array: object
        The array to check if it is a valid array-like

    Returns
    -------
    is_valid: bool
        True if and only if the given array is either a numpy array
        or a python sequence, in which the type entirely consists of
        numerical values, otherwise False.
    """
    # Check for valid numpy array
    if isinstance(array, np.ndarray):
        if array.size == 0:
            return True
        return (
            np.issubdtype(array.dtype, np.number)
            or np.issubdtype(array.dtype, np.floating)
            or np.issubdtype(array.dtype, bool)
        )

    # Check for numerical sequence
    if isinstance(array, Sequence) and not isinstance(array, str):
        if len(array) == 0:
            return True

        if isinstance(array[0], Sequence) and not isinstance(array[0], str):
            # Multivariate case
            n_attributes = len(array[0])
            return all(
                isinstance(sample, Sequence)
                and not isinstance(sample, str)
                and len(sample) == n_attributes
                and all(isinstance(item, (int, float)) for item in sample)
                for sample in array
            )
        else:
            # Univariate case
            return all(isinstance(item, (int, float)) for item in array)

    # Default case
    return False


def is_univariate(X: np.ndarray) -> bool:
    """
    Checks if the given array equals a univariate time series.

    Parameters
    ----------
    X: array-like of shape (n_samples, n_attributes)
        The time series data to check if it is univariate.

    Returns
    -------
    is_univariate: bool
        True if and only if the given time series has only one dimension,
        or if the second dimension of the time series is of size 1.
    """
    return get_dimension(X) == 1


def get_dimension(X: np.ndarray) -> int:
    """
    Get the dimension of the given array.

    Parameters
    ----------
    X: array-like of shape (n_samples, n_attributes)
        The time series data to get the dimension from

    Returns
    -------
    n_attributes: int
        The number of attributes in the given time series.
    """
    X = np.array(X)
    if len(X.shape) == 1:
        return 1
    else:
        return X.shape[1]



import numpy as np

from dtaianomaly.data.DataSet import DataSet
from dtaianomaly.data.LazyDataLoader import LazyDataLoader


class UCRLoader(LazyDataLoader):
    """
    Lazy dataloader for the UCR suite of anomaly detection data sets.

    This implementation expects the file names to contain the start and
    stop time stamps of the single anomaly in the time series as:
    ``*_<train-test-split>_<start>_<stop>.txt``.
    """

    def _load(self) -> DataSet:

        # Extract the meta-information from the name of the file
        [*_, train_test_split, start_anomaly, end_anomaly] = self.path.rstrip(
            ".txt"
        ).split("_")
        train_test_split = int(train_test_split)
        start_anomaly = int(start_anomaly)
        end_anomaly = int(end_anomaly)

        # Load time series
        X = np.loadtxt(self.path)
        X_train = X[:train_test_split]
        X_test = X[train_test_split:]

        # To ensure the file extensions gets ignored
        y = np.zeros(shape=X.shape[0], dtype=int)
        y[start_anomaly:end_anomaly] = 1
        y_test = y[train_test_split:]

        # Return a DataSet object
        return DataSet(X_test=X_test, y_test=y_test, X_train=X_train)



import abc
import os
from pathlib import Path
from typing import List, Type, Union

from dtaianomaly.data.DataSet import DataSet
from dtaianomaly.PrettyPrintable import PrettyPrintable


class LazyDataLoader(PrettyPrintable):
    """
    A lazy dataloader for anomaly detection workflows

    This is a data loading utility to point towards a specific data set
    (with `path`) and to load it at a later point in time during
    execution of a workflow.

    This way we limit memory usage and allow for virtually unlimited scaling
    of the number of data sets in a workflow.

    Parameters
    ----------
    path: str
        Path to the relevant data set.
    do_caching: bool, default=False
        Whether to cache the loaded data or not

    Attributes
    ----------
    cache_ : DataSet
        Cached version of the loaded data set. Only available if ``do_caching==True``
        and the data has been loaded before.

    Raises
    ------
    FileNotFoundError
        If the given path does not point to an existing file or directory.
    """

    path: str
    do_caching: bool
    cache_: DataSet

    def __init__(self, path: Union[str, Path], do_caching: bool = False):
        if not (Path(path).is_file() or Path(path).is_dir()):
            raise FileNotFoundError(f"No such file or directory: {path}")
        self.path = str(path)
        self.do_caching = do_caching

    def load(self) -> DataSet:
        """
        Load the dataset. If ``do_caching==True``, the loaded will be saved in the
        cache if no cache is available yet, and the cached data will be returned.

        Returns
        -------
        data_set: DataSet
            The loaded dataset.
        """
        if self.do_caching:
            if not hasattr(self, "cache_"):
                self.cache_ = self._load()
            return self.cache_
        else:
            return self._load()

    @abc.abstractmethod
    def _load(self) -> DataSet:
        """Abstract method to effectively load the data."""


def from_directory(
    directory: Union[str, Path], dataloader: Type[LazyDataLoader], **kwargs
) -> List[LazyDataLoader]:
    """
    Construct a `LazyDataLoader` instance for every file in the given `directory`

    Parameters
    ----------
    directory: str or Path
        Path to the directory in question
    dataloader: LazyDataLoader **object**
        Class object of the data loader, called for constructing
        each data loader instance
    **kwargs:
        Additional arguments to be passed to the dataloader

    Returns
    -------
    data_loaders: List[LazyDataLoader]
        A list of the initialized data loaders, one for each data set in the
        given directory.

    Raises
    ------
    FileNotFoundError
        If `directory` cannot be found
    """
    if not Path(directory).is_dir():
        raise FileNotFoundError(f"No such directory: {directory}")

    all_files = [
        os.path.join(directory, f)
        for f in os.listdir(directory)
        if os.path.isfile(os.path.join(directory, f))
        or os.path.isdir(os.path.join(directory, f))
    ]
    return [dataloader(file, **kwargs) for file in all_files]



from typing import Optional, Union

import numpy as np
from sklearn.cluster import KMeans
from sklearn.exceptions import NotFittedError

from dtaianomaly import utils
from dtaianomaly.anomaly_detection import BaseDetector, Supervision
from dtaianomaly.anomaly_detection.windowing_utils import (
    check_is_valid_window_size,
    compute_window_size,
    reverse_sliding_window,
    sliding_window,
)


class KMeansAnomalyDetector(BaseDetector):
    """
    Use KMeans clustering to detect anomalies.

    KMeans anomaly detector [yairi2001fault]_ first clusters the data using
    the KMeasn clustering algorithm. Next, for new data, the corresponding
    cluster is predicted, and the distance to the cluster centroid is computed.
    This distance corresponds to the decision scores of this anomaly detector:
    if an instance is far from the centroid, it is more anomalous. The input
    of KMeans clustering is a sliding window.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs:
        Arguments to be passed to KMeans clustering of scikit-learn anomaly detector

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    k_means_ : KMeans
        The KMeans clustering algorithm from scikit-learn

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import KMeansAnomalyDetector
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> kmeans_ad = KMeansAnomalyDetector(10).fit(x)
    >>> kmeans_ad.decision_function(x)
    array([0.50321076, 0.5753145 , 0.61938076, ..., 0.29794485, 0.30720306,
           0.29857479])

    References
    ----------
    .. [yairi2001fault] T. Yairi, Y. Kato, and K. Hori. Fault detection by mining association
       rules from house-keeping data. In proceedings of the 6th International Symposium on
       Artificial Intelligence, Robotics and Automation in Space, volume 18, page 21. Citeseer,
       2001.
    """

    window_size: Union[int, str]
    stride: int
    kwargs: dict
    window_size_: int
    k_means_: KMeans

    def __init__(self, window_size: Union[int, str], stride: int = 1, **kwargs):
        super().__init__(Supervision.UNSUPERVISED)

        check_is_valid_window_size(window_size)
        if not isinstance(stride, int) or isinstance(stride, bool):
            raise TypeError("`stride` should be an integer")
        if stride < 1:
            raise ValueError("`stride` should be strictly positive")

        self.window_size = window_size
        self.stride = stride
        self.kwargs = kwargs
        KMeans(**self.kwargs)  # Check if KMeans can be initialized

    def fit(
        self, X: np.ndarray, y: Optional[np.ndarray] = None, **kwargs
    ) -> "KMeansAnomalyDetector":
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")

        X = np.asarray(X)
        self.window_size_ = compute_window_size(X, self.window_size, **kwargs)
        self.k_means_ = KMeans(**self.kwargs)
        self.k_means_.fit(sliding_window(X, self.window_size_, self.stride))

        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        if not utils.is_valid_array_like(X):
            raise ValueError("Input must be numerical array-like")
        if not hasattr(self, "k_means_") or not hasattr(self, "window_size_"):
            raise NotFittedError("Call the fit function before making predictions!")

        X = np.asarray(X)
        sliding_windows = sliding_window(X, self.window_size_, self.stride)
        clusters = self.k_means_.predict(sliding_windows)
        distance_to_cluster_centers = np.linalg.norm(
            sliding_windows - self.k_means_.cluster_centers_[clusters], axis=1
        )
        decision_scores = reverse_sliding_window(
            distance_to_cluster_centers, self.window_size_, self.stride, X.shape[0]
        )

        return decision_scores



from pyod.models.copod import COPOD

from dtaianomaly.anomaly_detection.BaseDetector import Supervision
from dtaianomaly.anomaly_detection.PyODAnomalyDetector import PyODAnomalyDetector


class CopulaBasedOutlierDetector(PyODAnomalyDetector):
    """
    Copula-based outlier detector (COPD) algorithm.

    COPOD [li2020copod]_ is based on modeling multivariate data distributions
    using Copula models. Copula functions separate the marginal distributions
    from the dependency structure of a multivariate distribution. This allows
    a copula to describe the joint distribution over the features using only
    the independent marginals, offering high flexibility when modeling high
    dimensional data. Outliers are consequently detected by measuring the
    tail probabilities. COPOD is parameter-free because the copula function
    does not involve learning or stochastic training.

    Notes
    -----
    The COPOD detector inherets from :py:class:`~dtaianomaly.anomaly_detection.PyODAnomalyDetector`.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs
        Arguments to be passed to the PyOD COPOD detector.

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : COPOD
        A COPOD detector of PyOD

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import CopulaBasedOutlierDetector
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> copod = CopulaBasedOutlierDetector(10).fit(x)
    >>> copod.decision_function(x)
    array([ 9.90110663,  9.67868282,  9.51525285, ..., 25.00182389,
           24.60594424, 24.30393026])

    References
    ----------
    .. [li2020copod] Li, Z., Zhao, Y., Botta, N., Ionescu, C. and Hu, X.
       COPOD: Copula-Based Outlier Detection. IEEE International Conference
       on Data Mining (ICDM), 2020.
    """

    def _initialize_detector(self, **kwargs) -> COPOD:
        return COPOD(**kwargs)

    def _supervision(self) -> Supervision:
        return Supervision.UNSUPERVISED



from pyod.models.iforest import IForest

from dtaianomaly.anomaly_detection.BaseDetector import Supervision
from dtaianomaly.anomaly_detection.PyODAnomalyDetector import PyODAnomalyDetector


class IsolationForest(PyODAnomalyDetector):
    """
    Anomaly detector based on the Isolation Forest algorithm.

    The isolation forest [Liu2008isolation]_ generates random binary trees to
    split the data. If an instance requires fewer splits to isolate it from
    the other data, it is nearer to the root of the tree, and consequently
    receives a higher anomaly score.

    Notes
    -----
    The isolation forest inherets from :py:class:`~dtaianomaly.anomaly_detection.PyodAnomalyDetector`.

    Parameters
    ----------
    window_size: int or str
        The window size to use for extracting sliding windows from the time series. This
        value will be passed to :py:meth:`~dtaianomaly.anomaly_detection.compute_window_size`.
    stride: int, default=1
        The stride, i.e., the step size for extracting sliding windows from the time series.
    **kwargs
        Arguments to be passed to the PyOD isolation forest.

    Attributes
    ----------
    window_size_: int
        The effectively used window size for this anomaly detector
    pyod_detector_ : IForest
        An Isolation Forest detector of PyOD

    Examples
    --------
    >>> from dtaianomaly.anomaly_detection import IsolationForest
    >>> from dtaianomaly.data import demonstration_time_series
    >>> x, y = demonstration_time_series()
    >>> isolation_forest = IsolationForest(10).fit(x)
    >>> isolation_forest.decision_function(x)
    array([-0.02301142, -0.01266304, -0.00786237, ..., -0.04561172,
           -0.0420979 , -0.04414417])

    References
    ----------
    .. [Liu2008isolation] F. T. Liu, K. M. Ting and Z. -H. Zhou, "Isolation Forest,"
       2008 Eighth IEEE International Conference on Data Mining, Pisa, Italy, 2008,
       pp. 413-422, doi: `10.1109/ICDM.2008.17 <https://doi.org/10.1109/ICDM.2008.17>`_.
    """

    def _initialize_detector(self, **kwargs) -> IForest:
        return IForest(**kwargs)

    def _supervision(self):
        return Supervision.UNSUPERVISED



import json
import os
from itertools import chain

from dtaianomaly import anomaly_detection, data, evaluation, preprocessing, thresholding
from dtaianomaly.workflow import Workflow


def workflow_from_config(path: str, max_size: int = 1000000):
    """
    Construct a Workflow instance based on a JSON file. The file is
    first parsed, and then interpreted to obtain a :py:class:`~dtaianomaly.workflow.Workflow`

    Parameters
    ----------
    path: str
        Path to the config file in JSON format
    max_size: int, optional
        Maximal size of the config file in bytes. Defaults to 1 MB.

    Returns
    -------
    workflow: Workflow
        The parsed workflow from the given config file.

    Raises
    ------
    TypeError
        If the given path is not a string.
    FileNotFoundError
        If the given path does not correspond to an existing file.
    ValueError
        If the given path does not refer to a json file.
    """
    if not isinstance(path, str):
        raise TypeError("Path expects a string")
    if not os.path.exists(path):
        raise FileNotFoundError("The given path does not exist!")
    if not path.endswith(".json"):
        raise ValueError("The given path should be a json file!")

    with open(path, "r") as file:
        # Check file size
        file.seek(0, 2)
        file_size = file.tell()
        if file_size > max_size:
            raise ValueError(f"File size exceeds maximum size of {max_size} bytes")
        file.seek(0)

        # Parse actual JSON
        parsed_config = json.load(file)

    return interpret_config(parsed_config)


def interpret_config(config: dict):
    """
    Actual parsing/interpretation logic

    All the different `_interpret_*` functions below check the config
    for the corresponding `dtaianomaly` objects. These functions should
    be extended when the full package is extended.

    Parameters
    ----------
    config: dict
        The config to parse

    Returns
    -------
    Workflow
        Containing all the components specified in the config
    """
    if not isinstance(config, dict):
        raise TypeError("Input should be a dictionary")

    return Workflow(
        dataloaders=interpret_dataloaders(config),
        preprocessors=interpret_preprocessing(config),
        detectors=interpret_detectors(config),
        metrics=interpret_metrics(config),
        thresholds=interpret_thresholds(config),
        **interpret_additional_information(config),
    )


###################################################################
# THRESHOLDS
###################################################################


def interpret_thresholds(config):
    if "thresholds" not in config:
        return None

    threshold_config = config["thresholds"]
    if isinstance(threshold_config, list):
        return [threshold_entry(entry) for entry in threshold_config]
    else:
        return [
            threshold_entry(threshold_config),
        ]


def threshold_entry(entry):
    threshold_type = entry["type"]
    entry_without_type = {key: value for key, value in entry.items() if key != "type"}

    if threshold_type == "FixedCutoff":
        return thresholding.FixedCutoff(**entry_without_type)

    elif threshold_type == "ContaminationRate":
        return thresholding.ContaminationRate(**entry_without_type)

    elif threshold_type == "TopN":
        return thresholding.TopN(**entry_without_type)
    else:
        raise ValueError(f"Invalid threshold entry: {entry}")


###################################################################
# DATALOADERS
###################################################################


def interpret_dataloaders(config):
    if "dataloaders" not in config:
        raise ValueError("No `dataloaders` key in the config")

    data_config = config["dataloaders"]
    if isinstance(data_config, list):
        data_loaders = []
        for entry in data_config:
            new_loaders = data_entry(entry)
            if isinstance(new_loaders, list):
                data_loaders.extend(new_loaders)
            else:
                data_loaders.append(new_loaders)
        return data_loaders

    else:
        return [data_entry(data_config)]


def data_entry(entry):
    data_type = entry["type"]
    entry_without_type = {key: value for key, value in entry.items() if key != "type"}

    if data_type == "UCRLoader":
        return data.UCRLoader(**entry_without_type)

    elif data_type == "directory":
        if len(entry_without_type) != 2:
            raise ValueError(f"Incorrect number of items in entry: {entry}")
        if "path" not in entry:
            raise TypeError(f"Entry should have a path keyword: {entry}")
        if "base_type" not in entry:
            raise TypeError(f"Entry should have a base_type keyword: {entry}")

        if entry["base_type"] == "UCRLoader":
            base_type = data.UCRLoader
        else:
            raise ValueError(f"Invalid base type: {entry}")

        return data.from_directory(entry["path"], base_type)

    else:
        raise ValueError(f"Invalid data entry: {entry}")


###################################################################
# METRICS
###################################################################


def interpret_metrics(config):
    if "metrics" not in config:
        raise ValueError("No `metrics` key in the config")

    metric_config = config["metrics"]
    if isinstance(metric_config, list):
        return [metric_entry(entry) for entry in metric_config]
    else:
        return [metric_entry(metric_config)]


def metric_entry(entry):
    metric_type = entry["type"]
    entry_without_type = {key: value for key, value in entry.items() if key != "type"}

    if metric_type == "Precision":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return evaluation.Precision()

    elif metric_type == "Recall":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return evaluation.Recall()

    elif metric_type == "FBeta":
        return evaluation.FBeta(**entry_without_type)

    elif metric_type == "AreaUnderROC":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return evaluation.AreaUnderROC()

    elif metric_type == "AreaUnderPR":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return evaluation.AreaUnderPR()

    elif metric_type == "PointAdjustedPrecision":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return evaluation.PointAdjustedPrecision()

    elif metric_type == "PointAdjustedRecall":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return evaluation.PointAdjustedRecall()

    elif metric_type == "PointAdjustedFBeta":
        return evaluation.PointAdjustedFBeta(**entry_without_type)

    elif metric_type == "ThresholdMetric":
        if len(entry_without_type) != 2:
            raise TypeError(
                f"BestThresholdMetric must have thresholder and metric as key: {entry}"
            )
        if "thresholder" not in entry:
            raise ValueError(
                f"BestThresholdMetric must have thresholder as key: {entry}"
            )
        if "metric" not in entry:
            raise ValueError(f"BestThresholdMetric must have metric as key: {entry}")
        return evaluation.ThresholdMetric(
            thresholder=threshold_entry(entry["thresholder"]),
            metric=metric_entry(entry["metric"]),
        )

    elif metric_type == "BestThresholdMetric":
        if len(entry_without_type) != 1:
            raise TypeError(f"BestThresholdMetric must have metric as key: {entry}")
        if "metric" not in entry:
            raise ValueError(f"BestThresholdMetric must have metric as key: {entry}")
        return evaluation.BestThresholdMetric(metric=metric_entry(entry["metric"]))

    else:
        raise ValueError(f"Invalid metric entry: {entry}")


###################################################################
# DETECTORS
###################################################################


def interpret_detectors(config):
    if "detectors" not in config:
        raise ValueError("No `detectors` key in the config")

    detector_config = config["detectors"]
    if isinstance(detector_config, list):
        return [detector_entry(entry) for entry in detector_config]
    else:
        return [detector_entry(detector_config)]


def detector_entry(entry):
    detector_type = entry["type"]
    entry_without_type = {key: value for key, value in entry.items() if key != "type"}

    if detector_type == "AlwaysNormal":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return anomaly_detection.baselines.AlwaysNormal()

    elif detector_type == "AlwaysAnomalous":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return anomaly_detection.baselines.AlwaysAnomalous()

    elif detector_type == "RandomDetector":
        return anomaly_detection.baselines.RandomDetector(**entry_without_type)

    elif detector_type == "MatrixProfileDetector":
        return anomaly_detection.MatrixProfileDetector(**entry_without_type)

    elif detector_type == "IsolationForest":
        return anomaly_detection.IsolationForest(**entry_without_type)

    elif detector_type == "LocalOutlierFactor":
        return anomaly_detection.LocalOutlierFactor(**entry_without_type)

    elif detector_type == "MedianMethod":
        return anomaly_detection.MedianMethod(**entry_without_type)

    elif detector_type == "KNearestNeighbors":
        return anomaly_detection.KNearestNeighbors(**entry_without_type)

    elif detector_type == "HistogramBasedOutlierScore":
        return anomaly_detection.HistogramBasedOutlierScore(**entry_without_type)

    elif detector_type == "PrincipalComponentAnalysis":
        return anomaly_detection.PrincipalComponentAnalysis(**entry_without_type)

    elif detector_type == "KernelPrincipalComponentAnalysis":
        return anomaly_detection.KernelPrincipalComponentAnalysis(**entry_without_type)

    elif detector_type == "RobustPrincipalComponentAnalysis":
        return anomaly_detection.RobustPrincipalComponentAnalysis(**entry_without_type)

    elif detector_type == "OneClassSupportVectorMachine":
        return anomaly_detection.OneClassSupportVectorMachine(**entry_without_type)

    elif detector_type == "ClusterBasedLocalOutlierFactor":
        return anomaly_detection.ClusterBasedLocalOutlierFactor(**entry_without_type)

    elif detector_type == "KMeansAnomalyDetector":
        return anomaly_detection.KMeansAnomalyDetector(**entry_without_type)

    elif detector_type == "CopulaBasedOutlierDetector":
        return anomaly_detection.CopulaBasedOutlierDetector(**entry_without_type)

    else:
        raise ValueError(f"Invalid detector entry: {entry}")


###################################################################
# PREPROCESSING
###################################################################


def interpret_preprocessing(config):
    if "preprocessors" not in config:
        return None

    preprocessing_config = config["preprocessors"]

    if isinstance(preprocessing_config, list):
        return [preprocessing_entry(entry) for entry in preprocessing_config]
    else:
        return [preprocessing_entry(preprocessing_config)]


def preprocessing_entry(entry):
    processing_type = entry["type"]
    entry_without_type = {key: value for key, value in entry.items() if key != "type"}

    if processing_type == "Identity":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return preprocessing.Identity()

    elif processing_type == "MinMaxScaler":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return preprocessing.MinMaxScaler()

    elif processing_type == "StandardScaler":
        if len(entry_without_type) > 0:
            raise TypeError(f"Too many parameters given for entry: {entry}")
        return preprocessing.StandardScaler()

    elif processing_type == "MovingAverage":
        return preprocessing.MovingAverage(**entry_without_type)

    elif processing_type == "ExponentialMovingAverage":
        return preprocessing.ExponentialMovingAverage(**entry_without_type)

    elif processing_type == "NbSamplesUnderSampler":
        return preprocessing.NbSamplesUnderSampler(**entry_without_type)

    elif processing_type == "SamplingRateUnderSampler":
        return preprocessing.SamplingRateUnderSampler(**entry_without_type)

    elif processing_type == "Differencing":
        return preprocessing.Differencing(**entry_without_type)

    elif processing_type == "PiecewiseAggregateApproximation":
        return preprocessing.PiecewiseAggregateApproximation(**entry_without_type)

    elif processing_type == "RobustScaler":
        return preprocessing.RobustScaler(**entry_without_type)

    elif processing_type == "ChainedPreprocessor":
        if len(entry_without_type) != 1:
            raise TypeError(
                f"ChainedPreprocessor must have base_preprocessors as key: {entry}"
            )
        if "base_preprocessors" not in entry:
            raise ValueError(
                f"ChainedPreprocessor must have base_preprocessors as key: {entry}"
            )
        if not isinstance(entry["base_preprocessors"], list):
            raise ValueError(
                "A chained preprocessor should have a list as base_preprocessors!"
            )
        return preprocessing.ChainedPreprocessor(
            [preprocessing_entry(e) for e in entry["base_preprocessors"]]
        )

    else:
        raise ValueError(f"Invalid preprocessing config: {entry}")


###################################################################
# ADDITIONAL INFORMATION
###################################################################


def interpret_additional_information(config):
    return {
        feature: config[feature]
        for feature in ["n_jobs", "trace_memory"]
        if feature in config
    }
